/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ChatClient.h"
#include "Debug.h"
#include "IOString.h"
#include "XmlDef.h"
#include "XmlDocument.h"

#include <QTextStream>


//_______________________________________________________
int& ChatClient::_counter( void )
{
    static int counter( 0 );
    return counter;
}

//_______________________________________________________
ChatClient::ChatClient( QObject* parent, QTcpSocket* socket ):
    QObject( parent ),
    Counter( "ChatClient" ),
    id_( _counter()++ ),
    socket_( socket )
{
    Debug::Throw( "ChatClient::ChatClient.\n" );
    Q_CHECK_PTR( socket );
    connect( socket_, SIGNAL(connected()), SLOT(_sendCommands()) );
    connect( socket_, SIGNAL(readyRead()), SLOT(_readMessage()) );
}

//_______________________________________________________
ChatClient::~ChatClient( void )
{ Debug::Throw( "ChatClient::~ChatClient.\n" ); }

//_______________________________________________________
bool ChatClient::sendCommand( const ChatCommand& command )
{

    commands_ << command;
    if( socket_->state() ==  QAbstractSocket::ConnectedState ) _sendCommands();
    return true;

}

//_______________________________________________________
void ChatClient::_sendCommands( void )
{

    if( commands_.empty() ) return;

    QDomDocument document;
    QDomElement top = document.appendChild( document.createElement( Xml::Transmission ) ).toElement();
    while( commands_.size() && socket_->state() == QAbstractSocket::ConnectedState )
    {
        top.appendChild( commands_.front().domElement( document ) );
        commands_.removeFirst();
    }

    QTextStream os( socket_ );
    os << document.toString() << endl;

}

//_______________________________________________________
bool ChatClient::_readMessage( void )
{

    // read everything from socket and store as message
    IOString message( socket() );
    if( message.isEmpty() ) return false;

    // add to buffer
    buffer_.append( message );

    // parse buffer
    static const QString beginTag = QString( "<%1>" ).arg( Xml::Transmission );
    static const QString endTag = QString( "</%1>" ).arg( Xml::Transmission );

    forever
    {

        // get first tag
        int beginPosition( _messageBuffer().text().indexOf( beginTag, _messageBuffer().position() ) );
        if( beginPosition < 0 ) break;

        // get end tag
        int endPosition( _messageBuffer().text().indexOf( endTag, beginPosition+beginTag.size() ) );
        if( endPosition < 0 ) break;

        // create QDomDocument
        QString local( _messageBuffer().text().mid( beginPosition, endPosition+endTag.size()-beginPosition ) );

        // create document
        XmlDocument document;
        if( !document.setContent( local ) )
        {

            Debug::Throw() << document.error() << endl;

        } else {

            // parse document
            QDomElement docElement = document.documentElement();
            Q_ASSERT( docElement.tagName() == Xml::Transmission );

            for(QDomNode node = docElement.firstChild(); !node.isNull(); node = node.nextSibling() )
            {
                QDomElement element = node.toElement();
                if( element.isNull() ) continue;
                if( element.tagName() == Xml::Command ) {

                    ChatCommand command( element );
                    emit commandAvailable( command.setClientId( id() ) );

                } else { Debug::Throw(0) << "ChatCommand::_readMessage - unrecognized tagname: " << element.tagName() << endl; }

            }

        }

        // flush buffer
        _messageBuffer().flush( endPosition+endTag.size() );

    }

    return true;

}

//____________________________________________________
QTextStream& operator << (QTextStream& out, const ChatClient& client )
{
    out << "client id: " << client.id_ << " socket: " << client.socket_;
    return out;
}
