/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "Debug.h"
#include "ChatCommand.h"
#include "XmlDef.h"
#include "XmlCommandLineArguments.h"
#include "XmlTimeStamp.h"

//__________________________________________________________________
ChatCommand::CommandMap& ChatCommand::_commandNames( void )
{
    static CommandMap names;
    return names;
}

//_________________________________________
void ChatCommand::_initializeCommandNames( void ) const
{
    if( !_commandNames().empty() ) return;
    _commandNames().insert( None, "" );
    _commandNames().insert( Message, QObject::tr( "Message" ) );

    _commandNames().insert( ConnectionRequest, QObject::tr( "Connection requested" ) );
    _commandNames().insert( ConnectionAccepted, QObject::tr( "Connection accepted" ) );
    _commandNames().insert( ClientConnected, QObject::tr( "Client connected" ) );
    _commandNames().insert( ClientDisconnected, QObject::tr( "Client disconnected" ) );

    _commandNames().insert( FetchMessages, QObject::tr( "Fetch messages from server" ) );
    _commandNames().insert( FetchConnectedClients, QObject::tr( "Fetch connected clients" ) );
    _commandNames().insert( ConnectedClient, QObject::tr( "Connected client" ) );

    _commandNames().insert( PingServer, QObject::tr( "Ping server" ) );
    _commandNames().insert( PongServer, QObject::tr( "Server pong" ) );

    _commandNames().insert( PingClient, QObject::tr( "Ping client" ) );
    _commandNames().insert( PongClient, QObject::tr( "Client pong" ) );
    _commandNames().insert( ClearConnectedClients, QObject::tr( "Clear list of connected clients" ) );
    _commandNames().insert( ClearMessageList, QObject::tr( "Clear local message list" ) );
    _commandNames().insert( UserStateChanged, QObject::tr( "Change user status" ) );
    _commandNames().insert( StartMessageLog, QObject::tr( "Start sending messages" ) );
    _commandNames().insert( StopMessageLog, QObject::tr( "Finished sending messages" ) );
    _commandNames().insert( UserTyping, QObject::tr( "User typing" ) );
    _commandNames().insert( UsernameChanged, QObject::tr( "User name changed" ) );

}

//___________________________________________
ChatCommand::ChatCommand( const ConnectionId& id, const CommandType& command ):
    Counter( "ChatCommand" ),
    creationTime_( TimeStamp::now() ),
    id_( id ),
    command_( command )
{ Debug::Throw( "ChatCommand::ChatCommand.\n" ); }

//___________________________________________
ChatCommand::ChatCommand( const QDomElement& element ):
    Counter( "ChatCommand" )
{

    Debug::Throw( "ChatCommand::ChatCommand (dom).\n" );

    // parse attributes
    QDomNamedNodeMap attributes( element.attributes() );
    for( int i=0; i<attributes.count(); i++ )
    {
        QDomAttr attribute( attributes.item( i ).toAttr() );
        if( attribute.isNull() ) continue;
        if( attribute.name() == Xml::Type ) setCommand( (CommandType) attribute.value().toUInt() );
        else Debug::Throw(0) << "ChatCommand::ChatCommand - unrecognized attribute: " << attribute.name() << endl;
    }

    // loop over children
    // parse children elements
    for(QDomNode chidNode = element.firstChild(); !chidNode.isNull(); chidNode = chidNode.nextSibling() )
    {
        QDomElement childElement = chidNode.toElement();
        if( childElement.isNull() ) continue;
        QString tagName( childElement.tagName() );
        if( tagName == Xml::Id ) setId( ConnectionId( childElement ) );
        else if( tagName == Xml::CreationTime ) setCreationTime( XmlTimeStamp( childElement ) );
        else if( tagName == Xml::ReceptionTime ) setReceptionTime( XmlTimeStamp( childElement ) );
        else if( tagName == Xml::Arguments ) setArguments( XmlCommandLineArguments( childElement ) );
        else if( tagName == Xml::Text ) text_ = XmlString( childElement.text() ).toText();
    }

}

//__________________________________________________
QDomElement ChatCommand::domElement( QDomDocument& document ) const
{

    Debug::Throw( "ChatCommand::domElement.\n" );
    QDomElement out( document.createElement( Xml::Command ) );

    // type
    out.setAttribute( Xml::Type, QString::number( command() ) );

    // id
    out.appendChild( id_.domElement( document ) );

    // time stamp
    if( creationTime_.isValid() )
    { out.appendChild( XmlTimeStamp( creationTime_ ).domElement( Xml::CreationTime, document ) ); }

    // time stamp
    if( receptionTime_.isValid() )
    { out.appendChild( XmlTimeStamp( receptionTime_ ).domElement( Xml::ReceptionTime, document ) ); }

    // arguments
    if( !arguments().isEmpty() ) out.appendChild( XmlCommandLineArguments(arguments()).domElement( Xml::Arguments, document ) );

    // raw tex
    if( !text_.isEmpty() )
    {
        QString text( text_ );
        out.
            appendChild( document.createElement( Xml::Text ) ).
            appendChild( document.createTextNode( XmlString( text ).toXml() ) );
    }

    return out;

}
