#ifndef ChatCommand_h
#define ChatCommand_h

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ConnectionId.h"
#include "CommandLineArguments.h"
#include "Counter.h"
#include "Debug.h"
#include "TimeStamp.h"
#include "XmlString.h"

#include <QDomElement>
#include <QDomDocument>
#include <QString>
#include <QHash>

//_________________________________________
//* command to be send over the server
class ChatCommand:public Counter
{

    public:

    //* command type
    enum CommandType
    {
        None,
        Message,
        ConnectionRequest,
        ConnectionAccepted,
        FetchMessages,
        ClientConnected,
        ClientDisconnected,
        FetchConnectedClients,
        ConnectedClient,
        PingServer,
        PongServer,
        PingClient,
        PongClient,
        ClearConnectedClients,
        ClearMessageList,
        UserStateChanged,
        StartMessageLog,
        StopMessageLog,
        UserTyping,
        UsernameChanged
    };

    //* list
    using List = QList<ChatCommand>;

    //* constructor
    ChatCommand( const ConnectionId& id = ConnectionId(), const CommandType& command = None );

    //* constructor
    ChatCommand( const QDomElement& );

    //* equal-to operator
    bool operator == ( const ChatCommand& other ) const
    {
        return creationTime() == other.creationTime() &&
            id() == other.id() &&
            command() == other.command();
    }

    //* less-than operator
    bool operator < ( const ChatCommand& other ) const
    {
        if( creationTime() != other.creationTime() ) return creationTime() < other.creationTime();
        if( id() != other.id() ) return id() < other.id();

        if( command() != other.command() ) return command() < other.command();
        return false;
    }

    //*@name accessors
    //@{

    //* dom element
    QDomElement domElement( QDomDocument& ) const;

    //* validity
    bool isValid( void ) const
    { return id_.isValid() && command_ != None && creationTime_.isValid(); }

    //* true if this command is a child of argument
    /*! this is needed to have commands appear in Tree view */
    bool isChild( const ChatCommand& command ) const
    { return creationTime() >= command.creationTime() && id() == command.id(); }

    //* time stamp
    const TimeStamp& creationTime( void ) const
    { return creationTime_; }

    //* time stamp
    const TimeStamp& receptionTime( void ) const
    { return receptionTime_; }

    //* client id
    unsigned int clientId( void ) const
    { return clientId_; }

    //* connection id
    const ConnectionId& id( void ) const
    { return id_; }

    //* command
    CommandType command( void ) const
    { return command_; }

    //* command name
    QString commandName( void ) const
    {
        _initializeCommandNames();
        return _commandNames()[ command() ];
    }

    //* raw text
    const QString& text( void ) const
    { return text_; }

    //* arguments
    const CommandLineArguments& arguments( void ) const
    { return arguments_; }

    //@}

    //*@name modifiers
    //@{

    //* client id
    ChatCommand& setClientId( unsigned int value )
    {
        clientId_ = value;
        return *this;
    }

    //* time stamp
    void setCreationTime( const TimeStamp& value )
    { creationTime_ = value; }

    //* time stamp
    void setReceptionTime( const TimeStamp& value )
    { receptionTime_ = value; }

    //* connection id
    void setId( const ConnectionId& id )
    { id_ = id; }

    //* command
    void setCommand( const CommandType& command )
    { command_ = command; }

    //* raw text
    void setText( const QString& value )
    { text_ = value; }

    //* argument
    void setArguments( const CommandLineArguments& arguments )
    { arguments_ = arguments; }

    //@}

    private:

    //* command names
    using CommandMap = QHash<CommandType, QString >;

    //* command names
    static CommandMap& _commandNames( void );

    //* map command types to names
    void _initializeCommandNames( void ) const;

    //* time stamp
    TimeStamp creationTime_;

    //* time stamp
    TimeStamp receptionTime_;

    //* client id
    unsigned int clientId_ = 0;

    //* connection id
    ConnectionId id_;

    //* command
    CommandType command_ = None;

    //* arguments
    CommandLineArguments arguments_;

    //* raw text
    QString text_;

};

#endif
