/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ChatMessage.h"

#include "Debug.h"
#include "XmlDef.h"
#include "XmlTimeStamp.h"

//___________________________________________
ChatMessage::ChatMessage( const ConnectionId& id, const QString& text ):
    Counter( "ChatMessage" ),
    creationTime_( TimeStamp::now() ),
    id_( id ),
    text_( text ),
    blockRange_( qMakePair( 0, 0 ) )
{ Debug::Throw( "ChatMessage::ChatMessage.\n" ); }

//___________________________________________
ChatMessage::ChatMessage( const QDomElement& element ):
    Counter( "ChatMessage" ),
    blockRange_( qMakePair( 0, 0 ) )
{

    Debug::Throw( "ChatMessage::ChatMessage (dom).\n" );

    // loop over children
    // parse children elements
    for(QDomNode chidNode = element.firstChild(); !chidNode.isNull(); chidNode = chidNode.nextSibling() )
    {
        QDomElement childElement = chidNode.toElement();
        if( childElement.isNull() ) continue;
        QString tagName( childElement.tagName() );
        if( tagName == Xml::Id ) setId( ConnectionId( childElement ) );
        else if( tagName == Xml::CreationTime ) setCreationTime( XmlTimeStamp( childElement ) );
        else if( tagName == Xml::ReceptionTime ) setReceptionTime( XmlTimeStamp( childElement ) );
        else if( tagName == Xml::Text ) text_ = XmlString( childElement.text() ).toText();
    }

}

//__________________________________________________
ChatMessage::ChatMessage( const ChatCommand& command ):
    Counter( "ChatMessage" ),
    creationTime_( command.creationTime() ),
    receptionTime_( command.receptionTime() ),
    id_( command.id() ),
    text_( command.text() ),
    blockRange_( qMakePair( 0, 0 ) )
{ Q_ASSERT( command.command() == ChatCommand::Message ); }

//__________________________________________________
QDomElement ChatMessage::domElement( QDomDocument& document ) const
{

    Debug::Throw( "ChatMessage::domElement.\n" );
    QDomElement out( document.createElement( Xml::Message ) );

    // id
    out.appendChild( id_.domElement( document ) );

    // time stamp
    if( creationTime_.isValid() )
    { out.appendChild( XmlTimeStamp( creationTime_ ).domElement( Xml::CreationTime, document ) ); }

    // time stamp
    if( receptionTime_.isValid() )
    { out.appendChild( XmlTimeStamp( receptionTime_ ).domElement( Xml::ReceptionTime, document ) ); }

    // raw tex
    if( !text_.isEmpty() )
    {
        QString text( text_ );
        out.
            appendChild( document.createElement( Xml::Text ) ).
            appendChild( document.createTextNode( XmlString( text ).toXml() ) );
    }

    return out;

}

//__________________________________________________
ChatCommand ChatMessage::toChatCommand( void ) const
{
    ChatCommand command( id_, ChatCommand::Message );
    command.setCreationTime( creationTime_ );
    command.setReceptionTime( receptionTime_ );
    command.setText( text_ );
    return command;
}

//_______________________________________________
ChatMessage::Set::Set( const QDomElement& parent )
{

    for(QDomNode node = parent.firstChild(); !node.isNull(); node = node.nextSibling() )
    {
        QDomElement element = node.toElement();
        if( element.isNull() ) continue;

        // special options
        if( element.tagName() == Xml::Message )
        {

            ChatMessage record( element );
            if( record.isValid() )
            {

                if( !record.receptionTime().isValid() ) record.setReceptionTime( record.creationTime() );
                insert( record );

            } else Debug::Throw( "ChatMessage::Set::Set - attend to add empty record. Discarded.\n" );

        } else Debug::Throw() << "ChatMessage::Set::Set - unrecognized tag " << element.tagName() << endl;
    }

}

//_______________________________________________
QDomElement ChatMessage::Set::domElement( QDomDocument& parent ) const
{

    Debug::Throw( "ChatMessage::Set::domElement.\n" );

    // create main element
    QDomElement out( parent.createElement( Xml::MessageList ) );

    // loop over records
    foreach( const auto& message, *this )
    {

        // check message validity
        if( !message.isValid() )
        {
            Debug::Throw(0)
                << "ChatMessage::Set::domElement - attend to  write invalid message. Discarded"
                << " user: " << message.id().user()
                << " timestamp: " << message.creationTime().toString()
                << " text: " << message.text()
                << endl;

            continue;
        }

        // do not save empty messages
        if( message.text().isEmpty() ) continue;

        // append message to dom element
        out.appendChild( message.domElement( parent ) );
    }

    return out;

}

//_______________________________________________
bool ChatMessage::List::insert( const ChatMessage& message )
{

    // check if message can be inserted at end
    if( isEmpty() || last().receptionTime() <= message.receptionTime() )
    {
        append( message );
        return false;
    }

    // navigate list and insert before first item whose reception time stamp is larger
    // than this one
    for( iterator iter = begin(); iter != end(); ++iter )
    {
        if( message.receptionTime() < iter->receptionTime() )
        {
            QList<ChatMessage>::insert( iter, message );
            return true;
        }
    }

    // append if not done yet
    append( message );
    return false;
}

//_______________________________________________
bool ChatMessage::List::insertUnique( const ChatMessage& message )
{

    bool inserted( false );
    bool removed( false );
    for( iterator iter = begin(); iter != end(); ++iter )
    {
        // remove
        if( !removed && message == *iter )
        {
            removed = true;
            iter = erase( iter );
            if( iter == end() ) break;
        }

        // insert
        if( !inserted && message.receptionTime() < iter->receptionTime() )
        {
            inserted = true;
            iter = QList<ChatMessage>::insert( iter, message );
        }

        // stop when iter has been inserted and found
        if( inserted && removed ) return true;

    }

    // append if not done yet
    if( !inserted )
    {

        append( message );
        return removed;

    } else return true;

}

//____________________________________________________
QTextStream& operator << (QTextStream& out, const ChatMessage& message )
{
    out
        << "id: " << message.id_
        << " creation: " << message.creationTime_.toString()
        << " reception: " << message.receptionTime_.toString()
        << " block range: (" << message.blockRange_.first << "," << message.blockRange_.second << ")"
        << " text: " << message.text_;
    return out;
}
