/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ChatMessageListDialog.h"

#include "BaseIconNames.h"
#include "ChatMessage.h"
#include "ContextMenu.h"
#include "IconEngine.h"
#include "QuestionDialog.h"
#include "ServerInterface.h"
#include "TreeView.h"
#include "XmlOptions.h"

#include <QHeaderView>
#include <QLayout>
#include <QShortcut>

//______________________________________________________________________
ChatMessageListDialog::ChatMessageListDialog( QWidget* parent ):
    CustomDialog( parent, OkButton|CancelButton )
{

    Debug::Throw( "ChatMessageListDialog::ChatMessageListDialog.\n" );
    setOptionName( "CHAT_MESSAGE_LIST_DIALOG" );

    QHBoxLayout* hLayout( new QHBoxLayout() );
    hLayout->setSpacing(5);
    hLayout->setMargin(0);
    mainLayout().addLayout( hLayout );

    // setup list
    hLayout->addWidget( list_ = new TreeView( this ) );
    list_->setModel( &model_ );
    list_->setSortingEnabled( false );
    list_->setSelectionMode( QAbstractItemView::ContiguousSelection );
    list_->setOptionName( "CHAT_MESSAGE_LIST" );

    // buttons
    QVBoxLayout* vLayout = new QVBoxLayout();
    vLayout->setSpacing(5);
    vLayout->setMargin(0);
    hLayout->addLayout( vLayout );

    vLayout->addWidget( removeButton_ = new QPushButton( IconEngine::get( IconNames::Remove ), tr( "Remove" ), this ) );
    connect( removeButton_, SIGNAL(clicked()), SLOT(_remove()) );

    vLayout->addWidget( reloadButton_ = new QPushButton( IconEngine::get( IconNames::Reload ), tr( "Reload" ), this ) );
    connect( reloadButton_, SIGNAL(clicked()), SLOT(_reload()) );

    vLayout->addStretch( 1 );

    // actions
    QMenu* menu( new ContextMenu( list_ ) );
    addAction( removeAction_ = new QAction( IconEngine::get( IconNames::Remove ), tr( "Remove" ), this ) );
    connect( removeAction_, SIGNAL(triggered()), SLOT(_remove()) );
    removeAction_->setShortcut( QKeySequence::Delete );
    menu->addAction( removeAction_ );

    addAction( reloadAction_ = new QAction( IconEngine::get( IconNames::Reload ), tr( "Reload" ), this ) );
    connect( reloadAction_, SIGNAL(triggered()), SLOT(_reload()) );
    reloadAction_->setShortcut( QKeySequence::Refresh );

    menu->addAction( reloadAction_ );
    // updates
    connect( list_->selectionModel(), SIGNAL(selectionChanged(QItemSelection,QItemSelection)), SLOT(_updateButtons()) );
    connect( list_->selectionModel(), SIGNAL(currentChanged(QModelIndex,QModelIndex)), SLOT(_updateButtons()) );

    // permanent remove
    connect( &okButton(), SIGNAL(clicked()), SLOT(_removePermanently()) );

    // shortcuts
    cancelButton().setShortcut( Qt::Key_Escape );
    new QShortcut( QKeySequence::Close, this, SLOT(reject()) );
    new QShortcut( QKeySequence::Quit, this, SLOT(reject()) );

    _reload();
    _updateButtons();

}

//______________________________________________________________________
void ChatMessageListDialog::_reload( void )
{

    Debug::Throw( "DocumentClassManager::Dialog::_reload.\n" );

    // update model
    ServerInterface interface( this );
    model_.set( interface.messages().toList() );

    // clear list of removed messages
    removedMessages_.clear();

}

//______________________________________________________________________
void ChatMessageListDialog::_remove( void )
{
    Debug::Throw( "ChatMessageListDialog::_remove.\n" );

    // loop over selected items
    const ChatMessage::List removedMessages( model_.get( list_->selectionModel()->selectedRows() ) );

    if( removedMessages.empty() ) return;

    QuestionDialog* dialog( _removeMessagesQuestionDialog( removedMessages ) );

    // ask confirmation
    QString buffer = (removedMessages.size() == 1 ) ?
        tr( "Remove the selected message ?" ):
        QString( tr( "Remove the selected %1 messages ?" ) ).arg( removedMessages.size() );
    dialog->setText( buffer );

    // remove
    if( dialog->exec() )
    {
        removedMessages_.append( removedMessages );
        model_.remove( removedMessages );
    }

    // delete dialog and return
    dialog->deleteLater();

    return;

}

//______________________________________________________________________
void ChatMessageListDialog::_removePermanently( void )
{
    Debug::Throw( "ChatMessageListDialog::_removePermanently.\n" );

    // do nothing of there is no messages foreseen for removal
    if( removedMessages_.empty() ) return;

    // sort messages based on reception time
    qSort( removedMessages_.begin(), removedMessages_.end(), ChatMessage::SortReceptionTimeFTor() );

    QuestionDialog* dialog( _removeMessagesQuestionDialog( removedMessages_ ) );

    // ask confirmation
    QString buffer = (removedMessages_.size() == 1 ) ?
        tr( "Permanently delete the selected message ?" ):
        QString( tr( "Permanently delete the selected %1 messages ?" ) ).arg( removedMessages_.size() );
    dialog->setText( buffer );

    // remove
    if( dialog->exec() )
    {
        ServerInterface interface( this );
        interface.setMessages( ChatMessage::Set().fromList( model_.get() ) );
        interface.writeMessages();
    }

    // delete dialog and return
    dialog->deleteLater();

    return;

}

//______________________________________________________________________
void ChatMessageListDialog::_updateButtons( void )
{
    // loop over selected items
    bool removeEnabled( !model_.get( list_->selectionModel()->selectedRows() ).isEmpty() );
    removeAction_->setEnabled( removeEnabled );
    removeButton_->setEnabled( removeEnabled );
}

//______________________________________________________________________
QuestionDialog* ChatMessageListDialog::_removeMessagesQuestionDialog( const ChatMessage::List& messages )
{
    QuestionDialog* dialog = new QuestionDialog( this );

    // create details tree view
    TreeView* list = new TreeView();
    ChatMessageModel* model = new ChatMessageModel( dialog );
    list->setModel( model );
    model->set( messages );

    // assign
    dialog->setDetails( list );

    return dialog;
}
