/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ClientApplication.h"

#include "ChatMessage.h"
#include "ClientInterface.h"
#include "ClientConfigurationDialog.h"
#include "HistoryMenu.h"
#include "IdleTime.h"
#include "LineEditor.h"
#include "MainWindow.h"
#include "ProgressStatusBar.h"
#include "SshConnection.h"
#include "SshInterface.h"
#include "SystemTrayIcon.h"
#include "SystemNotifications.h"
#include "XmlOptions.h"

#include <QApplication>
#include <QTimer>

//____________________________________________
ClientApplication::ClientApplication( CommandLineArguments arguments ) :
    BaseApplication( 0, arguments ),
    Counter( "ClientApplication" )
{}

//____________________________________________
ClientApplication::~ClientApplication( void ) = default;

//____________________________________________
bool ClientApplication::realizeWidget( void )
{
    Debug::Throw( "ClientApplication::realizeWidget.\n" );

    // check if the method has already been called.
    if( !BaseApplication::realizeWidget() ) return false;

    // actions
    _installActions();

    // ssh interface
    sshInterface_ = new Ssh::Interface( this, commandLineParser( _arguments() ) );

    // tcp client interface
    interface_ = new ClientInterface( this, commandLineParser( _arguments() ) );

    // system notifications
    systemNotifications_ = new SystemNotifications( this, applicationName(), applicationIcon() );

    // main window
    mainWindow_ = new MainWindow();
    mainWindow_->setId( interface_->id() );

    // configuration
    connect( this, SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    emit configurationChanged();

    // center
    mainWindow_->centerOnDesktop();

    // show, unless specified by options
    if( !(
        XmlOptions::get().get<bool>( "USE_SYSTEMTRAY" ) &&
        QSystemTrayIcon::isSystemTrayAvailable() &&
        XmlOptions::get().get<bool>( "START_HIDDEN" ) ) )
    { _toggleMainWindowVisibility(); }

    // connections
    connect( mainWindow_, SIGNAL(newMessage(ChatMessage)), interface_, SLOT(sendMessage(ChatMessage)) );
    connect( mainWindow_, SIGNAL(newMessage(ChatMessage)), interface_, SLOT(pingServer()) );
    connect( mainWindow_, SIGNAL(newNotification(QString)), this, SLOT(_notify(QString)) );
    connect( mainWindow_, SIGNAL(userStateChanged(ConnectionId::UserState)), interface_, SLOT(updateUserState(ConnectionId::UserState)) );
    connect( mainWindow_, SIGNAL(userTyping()), interface_, SLOT(sendUserTyping()) );

    connect( mainWindow_, SIGNAL(requestDisconnection()), interface_, SLOT(disconnect()) );

    connect( &mainWindow_->fetchConnectedClientsAction(), SIGNAL(triggered()), interface_, SLOT(fetchConnectedClients()) );
    connect( &mainWindow_->pingServerAction(), SIGNAL(triggered()), interface_, SLOT(pingServerNotify()) );
    connect( &mainWindow_->pingClientsAction(), SIGNAL(triggered()), interface_, SLOT(pingClientsNotify()) );
    connect( &mainWindow_->reloadAction(), SIGNAL(triggered()), interface_, SLOT(fetchMessages()) );
    connect( &mainWindow_->historyMenu(), SIGNAL(requestMessages(HistoryMenu::Delay)), interface_, SLOT(fetchMessages(HistoryMenu::Delay)) );
    connect( &mainWindow_->idleTime(), SIGNAL(idleStateChanged(bool)), interface_, SLOT(updateIdleState(bool)) );

    connect( interface_, SIGNAL(disconnected()), sshInterface_,SLOT(disconnect()) );
    connect( interface_, SIGNAL(userNameChanged(ConnectionId)), mainWindow_, SLOT(setId(ConnectionId)) );
    connect( interface_, SIGNAL(connectionStateChanged(ConnectionId,ClientInterface::State)), mainWindow_, SLOT(updateConnectionState(ConnectionId,ClientInterface::State)));
    connect( interface_, SIGNAL(userStateChanged(ConnectionId)), mainWindow_, SLOT(updateUserState(ConnectionId)));
    connect( interface_, SIGNAL(userTyping(ConnectionId)), mainWindow_, SLOT(notifyUserTyping(ConnectionId)));
    connect( interface_, SIGNAL(clearConnectedClients()), mainWindow_, SLOT(clearConnectedClients()) );
    connect( interface_, SIGNAL(clearMessageList()), mainWindow_, SLOT(clearMessageList()) );
    connect( interface_, SIGNAL(messageRecieved(ChatMessage)), mainWindow_, SLOT(printMessage(ChatMessage)) );
    connect( interface_, SIGNAL(error(QString,bool)), mainWindow_, SLOT(printError(QString,bool)) );
    connect( interface_, SIGNAL(busyStateChanged(bool,int)), mainWindow_, SLOT(updateBusyState(bool,int)) );

    connect( &sshInterface_->connection(), SIGNAL(message(QString)), &mainWindow_->statusBar().label(0), SLOT(setTextAndUpdate(QString)) );

    // connect interfaces
    if( XmlOptions::get().get<bool>( "AUTO_CONNECT" ) )
    {
        if( XmlOptions::get().get<bool>( "USE_SSH" ) && !commandLineParser( _arguments() ).hasFlag( "--no-ssh" ) )
        {

            sshInterface_->connect();

        } else interface_->connect();
    }

    Debug::Throw( "ClientApplication::realizeWidget - done.\n" );
    return true;

}

//____________________________________________
CommandLineParser ClientApplication::commandLineParser( CommandLineArguments arguments, bool ignoreWarnings ) const
{
    Debug::Throw( "ClientApplication::commandLineParser.\n" );
    CommandLineParser out( BaseApplication::commandLineParser() );
    out.setGroup( CommandLineParser::applicationGroupName );
    out.registerFlag( CommandLineParser::Tag( "--no-ssh" ), tr( "disable internal SSH tunneling to connect to server" ) );
    out.registerOption( CommandLineParser::Tag( "--chat-host" ), tr( "string" ), tr( "use specified host for chat communication" ) );
    out.registerOption( CommandLineParser::Tag( "--chat-port" ), tr( "integer" ), tr( "use specified port for server communication" ) );
    out.registerOption( CommandLineParser::Tag( "--user", "-u" ), tr( "string" ), tr( "user name" ) );
    if( !arguments.isEmpty() ) out.parse( arguments, ignoreWarnings );
    return out;

}

//____________________________________________
void ClientApplication::usage( void ) const
{
    _usage( "Homechat" );
    commandLineParser().usage();
    return;
}

//_______________________________________________
void ClientApplication::_configuration( void )
{
    Debug::Throw( "ClientApplication::_configuration.\n" );
    emit saveConfiguration();
    ClientConfigurationDialog dialog;
    connect( &dialog, SIGNAL(configurationChanged()), SIGNAL(configurationChanged()) );
    dialog.centerOnWidget( qApp->activeWindow() );
    dialog.exec();
}

//________________________________________________
bool ClientApplication::_processCommand( Server::ServerCommand command )
{

    Debug::Throw( "ClientApplication::_processCommand.\n" );
    if( BaseApplication::_processCommand( command ) ) return true;
    if( command.command() == Server::ServerCommand::Raise )
    {
        mainWindow_->uniconify();
        return true;

    } else return false;

}

//_______________________________________________
void ClientApplication::_updateConfiguration( void )
{

    Debug::Throw( "ClientApplication::_updateConfiguration.\n" );

    // notifications
    useSystemNotifications_ = XmlOptions::get().get<bool>( "USE_NOTIFICATIONS" );

    // system tray icon
    if( XmlOptions::get().get<bool>( "USE_SYSTEMTRAY" ) )
    {

        qApp->setQuitOnLastWindowClosed( false );
        if( !systemTrayIcon_ )
        {

            systemTrayIcon_ = new SystemTrayIcon();
            connect( systemTrayIcon_, SIGNAL( activated(QSystemTrayIcon::ActivationReason) ), SLOT( _systemTrayActivated(QSystemTrayIcon::ActivationReason)) );
            connect( mainWindow_, SIGNAL(hasPendingMessages(bool)), systemTrayIcon_, SLOT(updatePendingMessages(bool)) );
            connect( interface_, SIGNAL(connectionStateChanged(ClientInterface::State)), systemTrayIcon_, SLOT(updateConnectionState(ClientInterface::State)));

        }

        systemTrayIcon_->show();

    } else {

        qApp->setQuitOnLastWindowClosed( true );
        if( systemTrayIcon_ ) systemTrayIcon_->deleteLater();
        systemTrayIcon_ = nullptr;

    }

    // ssh connection
    disconnect( mainWindow_, SIGNAL(requestConnection()), 0, 0 );
    if( XmlOptions::get().get<bool>( "USE_SSH" ) && !commandLineParser( _arguments() ).hasFlag( "--no-ssh" ) )
    {

        connect( mainWindow_, SIGNAL(requestConnection()), sshInterface_, SLOT(connect()) );
        connect( sshInterface_, SIGNAL(connected()), interface_, SLOT(connect()) );

    } else {

        disconnect( sshInterface_, SIGNAL(connected()), 0, 0 );
        connect( mainWindow_, SIGNAL(requestConnection()), interface_, SLOT(connect()) );

    }

}

//_______________________________________________
void ClientApplication::_notify( QString message )
{

    Debug::Throw( "ClientApplication::_notify.\n" );

    // do nothing if disabled
    if( !useSystemNotifications_ ) return;

    // do nothing if mainWindow is active
    if( mainWindow_->isActiveWindow() ) return;

    if( SystemNotifications::isSupported() )
    {

        // try system notifications
        systemNotifications_->processMessage( tr( "Homechat" ), message );

    } else if( QSystemTrayIcon::supportsMessages() && QSystemTrayIcon::isSystemTrayAvailable() && systemTrayIcon_ ) {

        // try system tray messages
        systemTrayIcon_->showMessage( "Homechat", message );

    }

}

//_______________________________________________
void ClientApplication::_systemTrayActivated( QSystemTrayIcon::ActivationReason reason )
{
    Debug::Throw( "ClientApplication::_systemTrayActivated.\n" );
    switch( reason )
    {

        default: break;

        case QSystemTrayIcon::Trigger:
        _toggleMainWindowVisibility();
        break;

    }
}


//________________________________________________
void ClientApplication::_toggleMainWindowVisibility( void )
{
    Debug::Throw( "ClientApplication::_toggleMainWindowVisibility.\n" );

    if( mainWindow_->isHidden() || mainWindow_->isMinimized() )
    {

        mainWindowAction_->setText( tr( "Minimize" ) );
        mainWindow_->uniconify();

    } else {

        mainWindowAction_->setText( tr( "Restore" ) );
        mainWindow_->hide();

    }

}

//_______________________________________________
void ClientApplication::_installActions( void )
{
    Debug::Throw( "ClientApplication::_installActions.\n" );

    // spell check
    mainWindowAction_ = new QAction( tr( "Restore" ), this );
    connect( mainWindowAction_, SIGNAL(triggered()), SLOT(_toggleMainWindowVisibility()) );

}
