/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ClientConfigurationDialog.h"

#include "Debug.h"
#include "GridLayout.h"
#include "IconEngine.h"
#include "IconNames.h"
#include "IconSize.h"
#include "OptionCheckBox.h"
#include "OptionComboBox.h"
#include "OptionLineEditor.h"
#include "OptionSpinBox.h"
#include "SshConnection.h"
#include "SystemNotifications.h"

#include <QGroupBox>
#include <QLabel>
#include <QLayout>
#include <QPushButton>
#include <QSystemTrayIcon>

//_________________________________________________________
ClientConfigurationDialog::ClientConfigurationDialog( QWidget* parent ):
    BaseConfigurationDialog( parent )
{

    Debug::Throw( "ClientConfigurationDialog::ClientConfigurationDialog.\n" );
    setWindowTitle( tr( "Configuration - Homechat" ) );

    // base configuration
    baseConfiguration( nullptr, Base);
    textEditConfiguration( nullptr, ParagraphHighlight|TextEditionFlags );

    QWidget *box;
    OptionLineEditor* editor;
    OptionCheckBox* checkbox;
    OptionSpinBox* spinbox;

    {
        // identity
        QWidget* page = &addPage( IconEngine::get( IconNames::PreferencesUser ), tr( "Identity" ) );
        page->layout()->addWidget( box = new QWidget( page ) );
        GridLayout* gridLayout = new GridLayout();
        gridLayout->setMaxCount(2);
        gridLayout->setMargin(0);
        gridLayout->setColumnAlignment( 0, Qt::AlignRight|Qt::AlignVCenter );
        box->setLayout( gridLayout );

        QLabel* label;
        gridLayout->addWidget( label = new QLabel( tr( "User name:" ), box ) );
        gridLayout->addWidget( editor = new OptionLineEditor( box, "USERNAME" ) );
        label->setBuddy( editor );
        addOptionWidget( editor );

        gridLayout->increment();
        gridLayout->addWidget( checkbox = new OptionCheckBox( tr( "Auto-connect at startup" ), box, "AUTO_CONNECT" ) );
        addOptionWidget( checkbox );
    }

    {
        // appearance
        QWidget* page = &addPage( IconEngine::get( IconNames::PreferencesAppearance ), tr( "Appearance" ), tr( "Application's appearance" ) );
        page->layout()->addWidget( box = new QWidget( page ) );
        box->setLayout( new QVBoxLayout() );
        box->layout()->setMargin(0);

        box->layout()->addWidget( checkbox = new OptionCheckBox( tr( "Show tooltips" ), page, "SHOW_TOOLTIPS" ) );
        addOptionWidget( checkbox );

        box->layout()->addWidget( checkbox = new OptionCheckBox( tr( "Use emoticons" ), page, "USE_EMOTICONS" ) );
        addOptionWidget( checkbox );

        box->layout()->addWidget( checkbox = new OptionCheckBox( tr( "Outline html links" ), page, "SHOW_LINKS" ) );
        addOptionWidget( checkbox );

        QHBoxLayout* hLayout = new QHBoxLayout();
        hLayout->setMargin( 0 );
        box->layout()->addItem( hLayout );

        QLabel* label;
        hLayout->addWidget( label = new QLabel( tr( "Emoticon icon size:" ), box ) );
        hLayout->addWidget( spinbox = new OptionSpinBox( box, "EMOTICON_SIZE" ) );
        spinbox->setMinimum(8);
        spinbox->setMaximum(IconSize::Huge);
        spinbox->setSuffix( tr( " px" ) );
        label->setBuddy( spinbox );
        addOptionWidget( spinbox );
    }

    {
        // behavior
        QWidget* page = &addPage( IconEngine::get( IconNames::PreferencesAppearance ), tr( "Behavior" ), tr( "Application's behavior" ) );
        page->layout()->addWidget( box = new QWidget( page ) );
        box->setLayout( new QVBoxLayout() );
        box->layout()->setMargin(0);

        OptionCheckBox* sysTrayCheckBox;
        box->layout()->addWidget( sysTrayCheckBox = new OptionCheckBox( tr( "Use system tray" ), box, "USE_SYSTEMTRAY" ) );
        addOptionWidget( sysTrayCheckBox );
        sysTrayCheckBox->setEnabled( QSystemTrayIcon::isSystemTrayAvailable() );

        if( SystemNotifications::isSupported() )
        {
            box->layout()->addWidget( checkbox = new OptionCheckBox( tr( "Use system notifications" ), box, "USE_NOTIFICATIONS" ) );
            addOptionWidget( checkbox );
        }

        box->layout()->addWidget( checkbox = new OptionCheckBox( tr( "Hide window on startup" ), box, "START_HIDDEN" ) );
        addOptionWidget( checkbox );
        checkbox->setEnabled( false );
        connect( sysTrayCheckBox, SIGNAL(toggled(bool)), checkbox, SLOT(setEnabled(bool)) );

        box->layout()->addWidget( checkbox = new OptionCheckBox( tr( "Show progress bar while downloading messages" ), page, "SHOW_FETCH_PROGRESS" ) );
        addOptionWidget( checkbox );

        // history span
        GridLayout* gridLayout = new GridLayout();
        gridLayout->setMaxCount(2);
        gridLayout->setMargin(0);
        gridLayout->setSpacing(5);
        gridLayout->setColumnAlignment( 0, Qt::AlignRight|Qt::AlignVCenter );
        box->layout()->addItem( gridLayout );

        QLabel* label;
        gridLayout->addWidget( label = new QLabel( tr( "Mark away when inactive for" ), box ) );
        gridLayout->addWidget( spinbox = new OptionSpinBox( box, "IDLE_TIMEOUT" ) );
        spinbox->setMinimum(1);
        spinbox->setMaximum(60);
        spinbox->setSuffix( tr( " minutes" ) );
        label->setBuddy( spinbox );
        addOptionWidget( spinbox );

        gridLayout->addWidget( label = new QLabel( tr( "Check connections to server every" ), box ) );
        gridLayout->addWidget( spinbox = new OptionSpinBox( box, "CHECK_CONNECTIONS_TIMEOUT" ) );
        spinbox->setMinimum(0);
        spinbox->setMaximum(60);
        spinbox->setSuffix( tr( " minutes" ) );
        spinbox->setSpecialValueText( tr( "Never" ) );
        label->setBuddy( spinbox );
        addOptionWidget( spinbox );

        OptionComboBox* comboBox;
        gridLayout->addWidget( label = new QLabel( tr( "At startup, fetch history for" ), box ) );
        gridLayout->addWidget( comboBox = new OptionComboBox( box, "DEFAULT_HISTORY_SPAN" ) );
        comboBox->setUseValue( false );

        // must be kept in sync with HistoryMenu::Delay enumeration
        comboBox->addItems( QStringList()
            << tr( "Today" )
            << tr( "Yesterday" )
            << tr( "One Week" )
            << tr( "Two Weeks" )
            << tr( "On Month" )
            << tr( "Six Months" )
            << tr( "One Year" )
            << tr( "From Beginning" ) );
        label->setBuddy( comboBox );
        addOptionWidget( comboBox );


    }

    {
        // server configuration
        QWidget* page = &addPage( IconEngine::get( IconNames::PreferencesServer ), tr( "Server" ), tr( "Server configuration" ) );
        GridLayout* gridLayout;
        QLabel* label;

        // use ssh
        if( Ssh::Connection::isSupported() )
        {
            OptionCheckBox* checkbox;
            page->layout()->addWidget( checkbox = new OptionCheckBox( tr( "Connect to server using ssh tunnel" ), box, "USE_SSH" ) );
            addOptionWidget( checkbox );

            // ssh connection
            page->layout()->addWidget( box = new QGroupBox( tr( "Ssh Connection Settings" ), page ) );
            gridLayout = new GridLayout();
            gridLayout->setMaxCount(4);
            gridLayout->setColumnAlignment( 0, Qt::AlignRight|Qt::AlignVCenter );
            box->setLayout( gridLayout );
            box->setEnabled( false );
            connect( checkbox, SIGNAL(toggled(bool)), box, SLOT(setEnabled(bool)) );

            // ssh host
            gridLayout->addWidget( label = new QLabel( tr( "Host:" ), box ) );
            gridLayout->addWidget( editor = new OptionLineEditor( box, "SSH_HOST" ) );
            label->setBuddy( editor );
            addOptionWidget( editor );

            gridLayout->addWidget( label = new QLabel( tr( "Port:" ), box ) );
            gridLayout->addWidget( spinbox = new OptionSpinBox( box, "SSH_PORT", false ) );
            spinbox->setMinimum(0);
            spinbox->setMaximum(10000);
            label->setBuddy( spinbox );
            addOptionWidget( spinbox );

            // ssh user
            gridLayout->addWidget( label = new QLabel( tr( "User:" ), box ) );
            gridLayout->addWidget( editor = new OptionLineEditor( box, "SSH_USERNAME" ), 1, 1, 1, 3 );
            label->setBuddy( editor );
            addOptionWidget( editor );

            // checkbox
            OptionCheckBox* passwordCheckBox;
            gridLayout->addWidget( passwordCheckBox = new OptionCheckBox( tr( "Remember password" ), box, "SSH_REMEMBER_PASSWORD" ), 2, 1, 1, 3 );
            addOptionWidget( passwordCheckBox );

        }

        // server connection
        page->layout()->addWidget( box = new QGroupBox( tr( "Server Connection Settings" ), page ) );
        gridLayout = new GridLayout();
        gridLayout->setMaxCount(4);
        gridLayout->setColumnAlignment( 0, Qt::AlignRight|Qt::AlignVCenter );
        box->setLayout( gridLayout );

        // host
        gridLayout->addWidget( label = new QLabel( tr( "Host:" ), box ) );
        gridLayout->addWidget( editor = new OptionLineEditor( box, "CHAT_HOST" ) );
        label->setBuddy( editor );
        addOptionWidget( editor );

        // port
        gridLayout->addWidget( label = new QLabel( tr( "Port:" ), box ) );
        gridLayout->addWidget( spinbox = new OptionSpinBox( box, "CHAT_PORT", false ) );
        spinbox->setMinimum(0);
        spinbox->setMaximum(10000);
        label->setBuddy( spinbox );
        addOptionWidget( spinbox );

        // timeout
        page->layout()->addWidget( box = new QGroupBox( tr( "Timeouts" ), page ) );
        gridLayout = new GridLayout();
        gridLayout->setMaxCount(2);
        gridLayout->setColumnAlignment( 0, Qt::AlignRight|Qt::AlignVCenter );
        box->setLayout( gridLayout );

        gridLayout->addWidget( label = new QLabel( tr( "Authorization timeout:" ), box ) );
        gridLayout->addWidget( spinbox = new OptionSpinBox( box, "AUTHORIZATION_TIMEOUT" ) );
        spinbox->setSuffix( tr( "s" ) );
        spinbox->setSpecialValueText( tr( " Never" ) );
        spinbox->setMinimum( 0 );
        spinbox->setMaximum( 60 );
        label->setBuddy( spinbox );
        addOptionWidget( spinbox );

        gridLayout->addWidget( label = new QLabel( tr( "Ping timeout:" ), box ) );
        gridLayout->addWidget( spinbox = new OptionSpinBox( box, "PING_TIMEOUT" ) );
        spinbox->setSuffix( tr( "s" ) );
        spinbox->setSpecialValueText( tr( " Never" ) );
        spinbox->setMinimum( 0 );
        spinbox->setMaximum( 60 );
        label->setBuddy( spinbox );
        addOptionWidget( spinbox );

    }

    // load initial configuration
    read();

}
