/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ClientInterface.h"

#include "Singleton.h"
#include "XmlOptions.h"

#include <QDateTime>
#include <QHostInfo>
#include <QTimerEvent>
#include <QTimer>

//____________________________________________
ClientInterface::ClientInterface( QObject* parent, const CommandLineParser& parser ):
    QObject( parent ),
    Counter( "ClientInterface" ),
    parser_( parser ),
    client_( new ChatClient( this, new QTcpSocket( this ) ) )

{

    // connections
    connect( &client_->socket(), SIGNAL(error(QAbstractSocket::SocketError)), SLOT(_error(QAbstractSocket::SocketError)) );
    connect( &client_->socket(), SIGNAL(disconnected()), SLOT(_disconnected()) );
    connect( &client_->socket(), SIGNAL(connected()), SLOT(_connected()) );
    connect( client_, SIGNAL(commandAvailable(ChatCommand)), SLOT(_processCommand(ChatCommand)) );

    // configuration
    connect( Singleton::get().application(), SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    _updateConfiguration();

}

//_____________________________________________________
void ClientInterface::connect( void )
{
    Debug::Throw() << "ClientInterface::connect - " << host_.toString() << ":" << port_ << endl;
    client_->socket().abort();
    client_->socket().connectToHost( host_, port_ );
}

//_____________________________________________________
void ClientInterface::disconnect( void )
{
    Debug::Throw("ClientInterface::disconnect.\n" );
    client_->socket().disconnectFromHost();
    _setConnectionState( Disconnected );
}

//_________________________________________________
void ClientInterface::updateIdleState( bool isIdle )
{
    Debug::Throw() << "ClientInterface::updateIdleState - idle: " << isIdle << endl;
    if( state_ != ClientInterface::Connected ) return;
    if( isIdle )
    {

        if( id_.userState() == ConnectionId::Online )
        { updateUserState( ConnectionId::Away ); }

    } else if( id_.userState() == ConnectionId::Away ) {

        updateUserState( ConnectionId::Online );

    }
}

//____________________________________________
void ClientInterface::updateUserState( ConnectionId::UserState state )
{
    Debug::Throw( "ClientInterface::updateUserState.\n" );
    if( state != id_.userState() )
    {
        // update id
        id_.setUserState( state );

        // also store in options
        if( !(state == ConnectionId::Away || state == ConnectionId::Offline ) )
        { XmlOptions::get().set<int>( QString( "USER_STATE_%1" ).arg( id_.user() ), state ); }

        // emit command
        client_->sendCommand( ChatCommand( id_, ChatCommand::UserStateChanged ) );
    }
}


//____________________________________________
void ClientInterface::pingServer( void )
{
    Debug::Throw( "ClientInterface::pingServer.\n" );
    if( !pingServerTimer_.isActive() && pingTimeout_ > 0 )
    {
        pingServerTimer_.start( pingTimeout_, this );
        client_->sendCommand( ChatCommand( id_, ChatCommand::PingServer ) );
    }
}

//____________________________________________
void ClientInterface::pingClients( void )
{
    Debug::Throw( "ClientInterface::pingClients.\n" );
    if( !pingClientTimer_.isActive() && pingTimeout_ > 0 )
    {

        // generate list of ids ping
        // do nothing if non found
        pingIds_ = connectedIds_;
        pingIds_.remove( id_ );
        if( !pingIds_.isEmpty() )
        {
            pingClientTimer_.start( pingTimeout_, this );
            client_->sendCommand( ChatCommand( id_, ChatCommand::PingClient ) );
        }

    }

}

//____________________________________________
void ClientInterface::pingServerNotify( void )
{
    Debug::Throw( "ClientInterface::pingServerNotify.\n" );
    notifyPingServerSuccess_ = true;
    pingServer();
}

//____________________________________________
void ClientInterface::pingClientsNotify( void )
{
    Debug::Throw( "ClientInterface::pingClientsNotify.\n" );
    notifyPingClientSuccess_ = true;
    pingClients();
}

//____________________________________________
void ClientInterface::fetchConnectedClients( void )
{
    Debug::Throw( "ClientInterface::fetchConnectedClients.\n" );
    client_->sendCommand( ChatCommand( id_, ChatCommand::FetchConnectedClients ) );
}

//____________________________________________
void ClientInterface::fetchMessages( void )
{
    Debug::Throw( "ClientInterface::fetchMessages.\n" );
    if( firstMessageTimeStamp_.isValid() )
    {

        ChatCommand command( id_, ChatCommand::FetchMessages );
        command.setArguments( QStringList() << command.commandName() << "-t" << QString::number( firstMessageTimeStamp_ ) );
        client_->sendCommand( command );


    } else {

        fetchMessages( (HistoryMenu::Delay) XmlOptions::get().get<int>( "DEFAULT_HISTORY_SPAN" ) );

    }
}

//____________________________________________
void ClientInterface::fetchMessages( HistoryMenu::Delay delay )
{
    Debug::Throw() << "ClientInterface::fetchMessages - delay: " << delay << endl;

    // reset first message time stamp
    firstMessageTimeStamp_ = TimeStamp();

    // create command
    ChatCommand command( id_, ChatCommand::FetchMessages );

    // define relevant fetch time
    TimeStamp timeStamp;
    switch( delay )
    {
        default: break;
        case HistoryMenu::AllMessages: break;

        case HistoryMenu::Today: timeStamp = QDateTime( QDate::currentDate() ).toTime_t(); break;
        case HistoryMenu::OneDay: timeStamp = QDateTime( QDate::currentDate() ).addDays(-1).toTime_t(); break;
        case HistoryMenu::OneWeek: timeStamp = QDateTime( QDate::currentDate() ).addDays(-7).toTime_t(); break;
        case HistoryMenu::TwoWeeks: timeStamp = QDateTime( QDate::currentDate() ).addDays(-14).toTime_t(); break;
        case HistoryMenu::OneMonth: timeStamp = QDateTime( QDate::currentDate() ).addMonths(-1).toTime_t(); break;
        case HistoryMenu::SixMonths: timeStamp = QDateTime( QDate::currentDate() ).addMonths(-6).toTime_t(); break;
        case HistoryMenu::OneYear: timeStamp = QDateTime( QDate::currentDate() ).addYears(-1).toTime_t(); break;
    }

    // add argument, if needed
    if( timeStamp.isValid() )
    { command.setArguments( QStringList() << command.commandName() << "-t" << QString::number( timeStamp ) ); }

    // send command
    client_->sendCommand( command );
}

//____________________________________________
void ClientInterface::sendUserTyping( void )
{ client_->sendCommand( ChatCommand( id_, ChatCommand::UserTyping ) ); }

//____________________________________________
void ClientInterface::sendMessage( ChatMessage message )
{
    Debug::Throw( "ClientInterface::sendMessage.\n" );
    if( state_ == Connected )
    {

        // send directly
        client_->sendCommand( message.toChatCommand() );

    } else {

        // append to pending messages
        ChatMessage::List::iterator iter( qFind( pendingMessages_.begin(), pendingMessages_.end(), message ) );
        if( iter != pendingMessages_.end() )
        {

            message.setReceptionTime( iter->receptionTime() );
            *iter = message;

        } else {

            if( !message.receptionTime().isValid() )
            { message.setReceptionTime( TimeStamp::now() ); }

            pendingMessages_.append( message );

        }
    }
}

//_____________________________________________________
void ClientInterface::timerEvent( QTimerEvent* event )
{
    Debug::Throw( "ClientInterface::timerEvent.\n" );
    if( event->timerId() == authorizationTimer_.timerId() )
    {

        authorizationTimer_.stop();
        if( state_ == AwaitingReply )
        {
            disconnect();
            emit error( tr( "Failed to get authorization from server" ), true );
            Debug::Throw( "ClientInterface::timerEvent - timeout while waiting for authorization.\n" );
        }

    } else if( event->timerId() == pingServerTimer_.timerId() ) {

        pingServerTimer_.stop();
        notifyPingServerSuccess_ = false;

        // failed to contact server
        if( state_ == Connected )
        {
            disconnect();
            emit error( tr( "Lost connection to server"), true );
            Debug::Throw( "ClientInterface::timerEvent - timeout while waiting for server reply.\n" );
        }

    } else if( event->timerId() == pingClientTimer_.timerId() ) {

        pingClientTimer_.stop();

        ConnectionId::Set failedIds;
        foreach( const auto& id, pingIds_ )
        { if( id.userState() != ConnectionId::Offline ) failedIds.insert( id ); }

        // mark failed ids as offline
        foreach( auto id, failedIds )
        {
            id.setUserState( ConnectionId::Offline );
            _updateUserState( id );
        }

        if( failedIds.size() == 1 ) {

            emit error( QString( tr( "Failed to contact user %1" ) ).arg( failedIds.begin()->user() ) );

        } else if( !failedIds.isEmpty() ) {

            QString buffer( tr( "The following users could not be contacted:\n" ) );
            foreach( const auto& id, failedIds )
            { buffer += QString( tr( " -%1\n" ) ).arg( id.user() ); }

            emit error( buffer );

        }

        pingIds_.clear();
        notifyPingClientSuccess_ = false;

    } else if( event->timerId() == checkConnectionsTimer_.timerId() ) {

        if( state_ == Connected )
        {
            pingServer();
            pingClients();
        }

    } else return QObject::timerEvent( event );
}

//_____________________________________________________
void ClientInterface::_error( QAbstractSocket::SocketError )
{
    Debug::Throw() << "ClientInterface::_error - error:" << client_->socket().errorString() << endl;
    disconnect();
    emit error( client_->socket().errorString(), true );
}

//_____________________________________________________
void ClientInterface::_connected( void )
{
    Debug::Throw( "ClientInterface::_connected.\n" );
    _setConnectionState( AwaitingReply );

    // send connection request and start timer
    if( !authorizationTimer_.isActive() && authorizationTimeout_ > 0 )
    { authorizationTimer_.start( authorizationTimeout_, this ); }

    client_->sendCommand( ChatCommand( id_, ChatCommand::ConnectionRequest ) );
}

//_____________________________________________________
void ClientInterface::_disconnected( void )
{
    Debug::Throw( "ClientInterface::_disconnected.\n" );
    _setConnectionState( Disconnected );
}

//____________________________________________
void ClientInterface::_processCommand( ChatCommand command )
{

    Debug::Throw() << "ClientInterface::_processCommand - command: " << command.commandName() << " user: " << command.id().user() << endl;

    switch( command.command() )
    {

        // connection accepted
        case ChatCommand::ConnectionAccepted:
        {
            if( command.id() == id_ )
            {
                authorizationTimer_.stop();
                _setConnectionState( Connected );

                // send pending commands
                foreach( const auto& message, pendingMessages_ )
                { client_->sendCommand( message.toChatCommand() ); }
                pendingMessages_.clear();

                // fetch users
                fetchConnectedClients();

                // fetch messages
                fetchMessages();
            }

            break;
        }

        // pong recieved from server
        case ChatCommand::PongServer:
        {
            if( pingServerTimer_.isActive() )
            {
                pingServerTimer_.stop();

                if( notifyPingServerSuccess_ )
                { emit error( tr( "Successfully established contact with server" ) ); }
                notifyPingServerSuccess_ = false;

                break;
            }
        }

        // ping request from other client
        case ChatCommand::PingClient:
        {
            client_->sendCommand( ChatCommand( id_, ChatCommand::PongClient ) );
            break;
        }

        // pong recieved from other client
        case ChatCommand::PongClient:
        {
            // update user state, insert in connected clients list
            _updateUserState( command.id() );

            if( pingClientTimer_.isActive() )
            {

                // remove from set
                pingIds_.remove( command.id() );

                // emit success
                if( pingIds_.isEmpty() )
                {

                    pingClientTimer_.stop();
                    if( notifyPingClientSuccess_ )
                    { emit error( tr( "Successfully established contact with all connected users" ) ); }
                    notifyPingClientSuccess_ = false;
                }

            }

            break;

        }

        // new client connected
        case ChatCommand::ClientConnected:
        {
            connectedIds_.insert( command.id() );
            emit connectionStateChanged( command.id(), Connected );
            break;
        }

        // client disconnected
        case ChatCommand::ClientDisconnected:
        {
            connectedIds_.remove( command.id() );
            emit connectionStateChanged( command.id(), Disconnected );
            break;
        }

        // user state changed
        case ChatCommand::UserStateChanged:
        {
            _updateUserState( command.id() );
            break;
        }

        // clear connected clients list
        case ChatCommand::ClearConnectedClients:
        {
            connectedIds_.clear();
            pingIds_.clear();
            emit clearConnectedClients();
            break;
        }

        // clear messages
        case ChatCommand::ClearMessageList:
        {
            emit clearMessageList();
            break;
        }

        // remote user is typing
        case ChatCommand::UserTyping:
        {
            emit userTyping( command.id() );
            break;
        }

        // new message recieved from client
        case ChatCommand::Message:
        {
            const ChatMessage message( command );

            // update first message timeStamp
            if( message.receptionTime().isValid() && (!firstMessageTimeStamp_.isValid() ||  message.receptionTime() < firstMessageTimeStamp_ ) )
            { firstMessageTimeStamp_ = message.receptionTime(); }

            // emit signal
            emit messageRecieved( message );
            break;
        }

        // start recieving message logs (busy state)
        case ChatCommand::StartMessageLog:
        {
            if( chatCommandLineParser_.parse( command.arguments() ).hasOption( "-s" ) )
            {

                emit busyStateChanged( true, chatCommandLineParser_.option( "-s" ).toInt() );

            } else emit busyStateChanged( true );
            break;

        }

        // finished recieving message logs (busy state)
        case ChatCommand::StopMessageLog: emit busyStateChanged( false ); break;

        default: break;

    }

    return;

}

//______________________________________
void ClientInterface::_updateConfiguration( void )
{
    Debug::Throw( "ClientInterface::_updateConfiguration.\n" );

    authorizationTimeout_ = 1000*XmlOptions::get().get<int>( "AUTHORIZATION_TIMEOUT" );
    pingTimeout_ = 1000*XmlOptions::get().get<int>( "PING_TIMEOUT" );

    if( parser_.hasOption( "--chat-host" ) )
    {

        // read from option
        QHostInfo hostInfo( QHostInfo::fromName( parser_.option( "--chat-host" ) ) );
        if( !hostInfo.addresses().isEmpty() ) host_ = hostInfo.addresses().front();

    } else {

        // read from option
        QHostInfo hostInfo( QHostInfo::fromName( XmlOptions::get().raw( "CHAT_HOST" ) ) );
        if( !hostInfo.addresses().isEmpty() ) host_ = hostInfo.addresses().front();

    }

    // port
    if( parser_.hasOption( "--chat-port" ) ) port_ = parser_.option( "--chat-port" ).toUInt();
    else  port_ = XmlOptions::get().get<unsigned int>( "CHAT_PORT" );

    // overwrite user from command line argument
    QString userName;
    if( parser_.hasOption( "--user" ) ) userName = parser_.option( "--user" );
    else if( XmlOptions::get().contains( "USERNAME" ) ) userName = XmlOptions::get().raw( "USERNAME" );

    if( !( userName.isEmpty() || id_.user() == userName ) )
    {

        id_.setUser( userName );
        emit userNameChanged( id_ );

        if( state_ == Connected )
        { client_->sendCommand( ChatCommand( id_, ChatCommand::UsernameChanged ) ); }

    }

    // adjust user status
    const QString userStateOptionName( QString( "USER_STATE_%1" ).arg( id_.user() ) );
    if( XmlOptions::get().contains( userStateOptionName ) )
    { id_.setUserState( (ConnectionId::UserState) XmlOptions::get().get<int>( userStateOptionName ) ); }

    // check connections daemon
    const int checkConnectionsTimeout( 1000*60*XmlOptions::get().get<int>( "CHECK_CONNECTIONS_TIMEOUT" ) );
    if( checkConnectionsTimer_.isActive() ) checkConnectionsTimer_.stop();
    if( checkConnectionsTimeout > 0 ) checkConnectionsTimer_.start( checkConnectionsTimeout, this );

    return;
}

//______________________________________
void ClientInterface::_setConnectionState( State state )
{
    if( state != state_ )
    {
        state_ = state;
        emit connectionStateChanged( id_, state );
        emit connectionStateChanged( state );

        if( state == Disconnected ) emit disconnected();

    }
}

//______________________________________
void ClientInterface::_updateUserState( const ConnectionId& id )
{

    if( id == id_ ) return;

    auto iter = connectedIds_.find( id );
    if( iter != connectedIds_.end() )
    {

        ConnectionId& local( const_cast<ConnectionId&>( *iter ) );
        if( local.userState() != id.userState() )
        { local.setUserState( id.userState() ); }

    }

    emit userStateChanged( id );
}
