#ifndef ClientInterface_h
#define ClientInterface_h

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ChatClient.h"
#include "ChatCommandLineParser.h"
#include "ConnectionId.h"
#include "ChatMessage.h"
#include "HistoryMenu.h"

#include <QBasicTimer>
#include <QTcpServer>
#include <QHostAddress>

class ClientInterface: public QObject, public Counter
{

    //* Qt meta object declaration
    Q_OBJECT

    public:

    //* constructor
    ClientInterface( QObject*, const CommandLineParser& );

    //* destructor
    virtual ~ClientInterface( void ) = default;

    //* application state enumeration
    enum State
    {
        AwaitingReply,
        Connected,
        Disconnected
    };

    //*@name accessors
    //@{

    //* id
    const ConnectionId& id( void ) const
    { return id_; }

    //@}

    using QObject::connect;

    Q_SIGNALS:

    //* new error recieved
    void error( QString, bool tryReconnect = false );

    //* client connection state changed
    void connectionStateChanged( ConnectionId, ClientInterface::State );

    //* this client connection state changed
    void connectionStateChanged( ClientInterface::State );

    //* this client has disconnected
    void disconnected( void );

    //* user name has changed
    void userNameChanged( ConnectionId );

    //* user state changed
    void userStateChanged( ConnectionId );

    //* busy state changed
    void busyStateChanged( bool, int = 0 );

    //* user typing
    void userTyping( ConnectionId );

    //* clear list of connected clients
    void clearConnectedClients( void );

    //* clear message list
    void clearMessageList( void );

    //* new message recieved
    void messageRecieved( ChatMessage );

    public Q_SLOTS:

    //* connect
    virtual void connect( void );

    //* disconnect
    virtual void disconnect( void );

    //* update idle state
    virtual void updateIdleState( bool );

    //* update user state
    virtual void updateUserState( ConnectionId::UserState );

    //* ping server
    virtual void pingServer( void );

    //* ping clients
    virtual void pingClients( void );

    //* ping server
    virtual void pingServerNotify( void );

    //* ping clients
    virtual void pingClientsNotify( void );

    //* fetch connected clients
    virtual void fetchConnectedClients( void );

    //* fetch messages
    virtual void fetchMessages( void );

    //* fetch messages
    virtual void fetchMessages( HistoryMenu::Delay );

    //* notify that user is typing
    virtual void sendUserTyping( void );

    //* send message
    virtual void sendMessage( ChatMessage );

    protected:

    virtual void timerEvent( QTimerEvent* );

    protected Q_SLOTS:

    //* client recieves errors
    virtual void _error( QAbstractSocket::SocketError );

    //* connection granted
    virtual void _connected( void );

    //* connection closed
    virtual void _disconnected( void );

    //* process command
    virtual void _processCommand( ChatCommand );

    private Q_SLOTS:

    //* update configuration
    void _updateConfiguration( void );

    private:

    //* changes application state, emit signal if changed
    void _setConnectionState( State );

    //* changes user state
    void _updateUserState( const ConnectionId& );

    // command line parser (passed at initialization)
    CommandLineParser parser_;

    // host
    QHostAddress host_ = QHostAddress::Null;

    // port
    int port_ = 0;

    //* Client
    ChatClient* client_ = nullptr;

    //* chat command line parser
    ChatCommandLineParser chatCommandLineParser_;

    //* connection id
    ConnectionId id_;
    bool forceUsername_ = false;

    //* set of connected clients
    ConnectionId::Set connectedIds_;

    //* set of ping ids
    ConnectionId::Set pingIds_;

    //* manager status
    State state_ = Disconnected;

    //* pending messages, sent while connection is down
    ChatMessage::List pendingMessages_;

    //* reception timestamp of the oldest recieved message
    TimeStamp firstMessageTimeStamp_;

    //* timeout
    QBasicTimer authorizationTimer_;
    QBasicTimer pingServerTimer_;
    QBasicTimer pingClientTimer_;
    QBasicTimer checkConnectionsTimer_;

    int authorizationTimeout_ = 0;
    int pingTimeout_ = 0;
    bool notifyPingServerSuccess_ = false;
    bool notifyPingClientSuccess_ = false;

};

#endif
