/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ConnectedClientToolTipWidget.h"
#include "Debug.h"
#include "GridLayout.h"
#include "GridLayoutItem.h"
#include "QtUtil.h"
#include "Singleton.h"
#include "TimeStamp.h"
#include "XmlOptions.h"

#include <QLayout>

//_______________________________________________________
ConnectedClientToolTipWidget::ConnectedClientToolTipWidget( QWidget* parent ):
    BaseToolTipWidget( parent )
{

    Debug::Throw( "ConnectedClientToolTipWidget::ConnectedClientToolTipWidget.\n" );

    // layout
    QHBoxLayout* hLayout = new QHBoxLayout();
    hLayout->setMargin( 10 );
    hLayout->setSpacing( 10 );
    setLayout( hLayout );

    hLayout->addWidget( iconLabel_ = new QLabel( this ) );
    iconLabel_->setAlignment( Qt::AlignHCenter|Qt::AlignTop );

    QVBoxLayout* vLayout = new QVBoxLayout();
    vLayout->setMargin( 0 );
    vLayout->setSpacing( 5 );
    hLayout->addLayout( vLayout );

    // file
    vLayout->addWidget( userNameLabel_ = new QLabel( this ) );
    userNameLabel_->setAlignment( Qt::AlignCenter );
    userNameLabel_->setFont( QtUtil::titleFont( userNameLabel_->font() ) );
    userNameLabel_->setMargin( 1 );

    // separator
    vLayout->addWidget( separator_ = new QFrame( this ) );
    separator_->setFrameStyle( QFrame::HLine );

    // grid layout
    GridLayout* gridLayout = new GridLayout();
    gridLayout->setMaxCount( 2 );
    gridLayout->setColumnAlignment( 0, Qt::AlignVCenter|Qt::AlignRight );
    gridLayout->setColumnAlignment( 1, Qt::AlignVCenter|Qt::AlignLeft );
    gridLayout->setMargin( 0 );
    gridLayout->setSpacing( 5 );
    vLayout->addLayout( gridLayout );

    // items
    ( connectionTimeItem_ = new GridLayoutItem( this, gridLayout ) )->setKey( tr( "Online since:" ) );
    ( disconnectionTimeItem_ = new GridLayoutItem( this, gridLayout ) )->setKey( tr( "Last seen:" ) );

    // add stretch
    // vLayout->addStretch( 1 );

    // configuration
    connect( Singleton::get().application(), SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    _updateConfiguration();

}

//_______________________________________________________
void ConnectedClientToolTipWidget::setId( const ConnectionId& id, QIcon icon )
{
    Debug::Throw( "ConnectedClientToolTipWidget::setId.\n" );

    // local storage
    // icon_ = icon;
    id_ = id;

    // update icon
    if( !icon_.isNull() )
    {

        iconLabel_->setPixmap( icon_.pixmap( QSize( pixmapSize_, pixmapSize_ ) ) );
        iconLabel_->show();

    } else iconLabel_->hide();

    if( !id.user().isEmpty() )
    {

        userNameLabel_->setText( id.userString() );
        if( id.userState() == ConnectionId::Offline )
        {

            if( id.disconnectionTime().isValid() )
            {

                separator_->show();
                disconnectionTimeItem_->setText( _timeString( id.disconnectionTime() ) );

            } else {

                separator_->hide();
                disconnectionTimeItem_->hide();

            }

            connectionTimeItem_->hide();

        } else {

            if( id.connectionTime().isValid() )
            {

                separator_->show();
                connectionTimeItem_->setText( _timeString( id.connectionTime() ) );

            } else {

                separator_->hide();
                connectionTimeItem_->hide();

            }

            disconnectionTimeItem_->hide();

        }

    } else {

        // file and separator
        userNameLabel_->hide();
        separator_->hide();
        connectionTimeItem_->hide();
        disconnectionTimeItem_->hide();

    }

    adjustSize();

}

//_____________________________________________
void ConnectedClientToolTipWidget::_updateConfiguration( void )
{
    Debug::Throw( "ConnectedClientToolTipWidget::_updateConfiguration.\n" );
    if( XmlOptions::get().contains( "TOOLTIPS_PIXMAP_SIZE" ) ) setPixmapSize( XmlOptions::get().get<unsigned int>( "TOOLTIPS_PIXMAP_SIZE" ) );
}

//_____________________________________________
QString ConnectedClientToolTipWidget::_timeString( const TimeStamp& timeStamp ) const
{
    if( timeStamp.isSameDay( TimeStamp::now() ) )
    {

        return timeStamp.toString( TimeStamp::Time );

    } else return timeStamp.toString( TimeStamp::Short );

}
