/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ConnectionId.h"
#include "Debug.h"
#include "XmlDef.h"
#include "XmlTimeStamp.h"
#include "Util.h"

#include <QHostAddress>

//____________________________________________________
ConnectionId::ConnectionId( const QString& user, const QString& host ):
    Counter( "ConnectionId" ),
    user_( user ),
    host_( host )
{
    Debug::Throw( "ConnectionId::ConnectionId.\n" );
    if( user_.isNull() ) user_ = Util::user();
    if( host_.isNull() ) host_ = QHostAddress( QHostAddress::LocalHost ).toString();
}

//____________________________________________________
ConnectionId::ConnectionId( const QDomElement& element ):
    Counter( "ConnectionId" )
{
    Debug::Throw( "ConnectionId::ConnectionId (dom).\n" );

    // parse attributes
    QDomNamedNodeMap attributes( element.attributes() );
    for( int i=0; i<attributes.count(); i++ )
    {
        QDomAttr attribute( attributes.item( i ).toAttr() );
        if( attribute.isNull() ) continue;
        if( attribute.name() == Xml::User ) setUser( attribute.value() );
        else if( attribute.name() == Xml::Host ) setHost( attribute.value() );
        else if( attribute.name() == Xml::UserState ) setUserState( (UserState) attribute.value().toInt() );
        else Debug::Throw() << "ConnectionId::ConnectionId - unrecognized attribute " << attribute.name() << endl;
    }

    // loop over children
    // parse children elements
    for(QDomNode chidNode = element.firstChild(); !chidNode.isNull(); chidNode = chidNode.nextSibling() )
    {
        QDomElement childElement = chidNode.toElement();
        if( childElement.isNull() ) continue;
        QString tagName( childElement.tagName() );
        if( tagName == Xml::ConnectionTime ) setConnectionTime( XmlTimeStamp( childElement ) );
        else if( tagName == Xml::DisconnectionTime ) setDisconnectionTime( XmlTimeStamp( childElement ) );
        else Debug::Throw() << "ConnectionId::ConnectionId - unrecognized tag " << tagName << endl;
    }
}

//____________________________________________________
QDomElement ConnectionId::domElement( QDomDocument& document ) const
{

    Debug::Throw( "ConnectionId::domElement" );
    QDomElement out( document.createElement( Xml::Id ) );
    if( !user_.isEmpty() ) out.setAttribute( Xml::User, user_ );
    if( !host_.isEmpty() ) out.setAttribute( Xml::Host, host_ );
    out.setAttribute( Xml::UserState, QString::number( userState_ ) );

    // time stamp
    if( connectionTime_.isValid() )
    { out.appendChild( XmlTimeStamp( connectionTime_ ).domElement( Xml::ConnectionTime, document ) ); }

    // time stamp
    if( disconnectionTime_.isValid() )
    { out.appendChild( XmlTimeStamp( disconnectionTime_ ).domElement( Xml::DisconnectionTime, document ) ); }

    return out;

}

//____________________________________________________
QString ConnectionId::userString( void ) const
{
    QString user( user_ );
    if(!user.isEmpty() ) user[0] = user[0].toUpper();
    return user;
}

//____________________________________________________
QTextStream& operator << (QTextStream& out, const ConnectionId& id )
{
    out
        << "user: " << id.user()
        << " state: " << id.userState_
        << " connection: " << id.connectionTime_.toString()
        << " disconnection: " << id.disconnectionTime_.toString();
    return out;


}
