/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "EmoticonProvider.h"
#include "Debug.h"
#include "IconSize.h"

#include <QIcon>
#include <QStringList>
#include <QTextBlock>
#include <QTextDocumentFragment>
#include <QUrl>
#include <QSet>

//________________________________________________________________________
EmoticonProvider::EmoticonProvider( QObject* parent ):
    QObject( parent ),
    Counter( "EmoticonProvider" )
{
    Debug::Throw( "EmoticonProvider::EmoticonProvider.\n" );

    // install emoticons
    Emoticon::List emoticons;
    emoticons.append( Emoticon( QRegExp("\\:-{0,1}\\)"), QStringList() << ":-)" << ":)", ":/smile.svgz", tr( "Smile" ) ) );
    emoticons.append( Emoticon( QRegExp("\\:-{0,1}\\("), QStringList() << ":-(" << ":(", ":/sad.svgz", tr( "Sad" ) ) );
    emoticons.append( Emoticon( QRegExp("\\:-{0,1}D"), QStringList() << ":-D" << ":D", ":/big-smile.svgz", tr( "Big smile" ) ) );
    emoticons.append( Emoticon( QRegExp("\\:-{0,1}P"), QStringList() << ":-P" << ":P", ":/tongue.svgz", tr( "Smile with tongue out" ) ) );
    emoticons.append( Emoticon( QRegExp("\\:-{0,1}\\|"), QStringList() << ":-|" << ":|", ":/speechless.svgz", tr( "Speechless" ) ) );
    emoticons.append( Emoticon( QRegExp("\\:-{0,1}S"), QStringList() << ":-S" << ":S", ":/mixed.svgz", tr( "Mixed" ) ) );
    emoticons.append( Emoticon( QRegExp("\\;-{0,1}\\)"), QStringList() << ";-)" << ";)", ":/wink.svgz", tr( "Wink" ) ) );
    emoticons.append( Emoticon( QRegExp("\\:-{0,1}\\*"), QStringList() << ":-*" << ":*", ":/kiss.svgz", tr( "Kiss" ) ) );
    emoticons.append( Emoticon( QRegExp("(?:\\:'|;)\\("), QStringList() << ":'(" << ";(", ":/cry.svgz", tr( "Cry" ) ) );
    emoticons.append( Emoticon( QRegExp("\\:-{0,1}(?:o|O)"), QStringList() << ":-o" << ":o" << ":-O" << ":O", ":/surprised.svgz", tr( "Surprise" ) ) );
    emoticons.append( Emoticon( QRegExp("(?:x|X)-{0,1}\\("), QStringList() << "x-(" << "x(" << "X-(" << "X(", ":/angry.svgz", tr( "Angry" ) ) );
    emoticons.append( Emoticon( QRegExp("\\:-{0,1}(?:x|X)"), QStringList() << ":-x" << ":x" << ":-X" << ":X", ":/silent.svgz", tr( "Silent" ) ) );
    emoticons.append( Emoticon( QRegExp("B-{0,1}\\)"), QStringList() << "B-)" << "B)", ":/cool.svgz", tr( "Cool" ) ) );
    emoticons.append( Emoticon( QRegExp("<3"), QStringList() << "<3", ":/heart.svgz", tr( "Love" ) ) );

    // insert in hash
    QSet<QString> regexps;
    foreach( const auto& emoticon, emoticons )
    {
        if( !emoticon.isValid() ) continue;

        // append to list
        emoticons_.append( emoticon );

        // append to map
        foreach( const auto& string, emoticon.strings() )
        { emoticonMap_.insert( string, emoticon ); }

        // get regexp
        regexps.insert( QString( "(?:%1)" ).arg( emoticon.regExp().pattern() ) );

    }

    // build master regexp
    regexp_ = QRegExp( QString( "(%1)" ).arg( QStringList( regexps.toList() ).join( "|" ) ) );

}

//________________________________________________________________________
bool EmoticonProvider::setIconSize( int size )
{
    if( size == iconSize_ ) return false;

    iconSize_ = size;
    emoticonMap_.clear();

    // update all emoticons and re-insert in map
    for( auto iter = emoticons_.begin(); iter != emoticons_.end(); ++iter )
    {
        iter->setIconSize( QSize( size, size ) );

        // append to map
        foreach( const auto& string, iter->strings() )
        { emoticonMap_.insert( string, *iter ); }

    }

    return true;

}

//________________________________________________________________________
void EmoticonProvider::installResources( QTextDocument* document ) const
{
    Debug::Throw( "EmoticonProvider::install.\n" );
    foreach( const auto& emoticon, emoticons_ )
    { document->addResource(QTextDocument::ImageResource, QUrl( _resourceName( emoticon ) ), QVariant(emoticon.image() ) ); }
}

//________________________________________________________________________
void EmoticonProvider::process( QTextCursor cursor )
{

    if( !( enabled_ && iconSize_ > 0 ) ) return;

    int position( 0 );
    QString text( cursor.block().text() );

    while( ( position = text.indexOf( regexp_, position ) ) >= 0 )
    {
        // find emoticon matching regexp
        const int length( regexp_.matchedLength() );
        const auto iter( emoticonMap_.find( regexp_.cap(0) ) );
        if( iter != emoticonMap_.end() )
        {

            // set cursor at the right position
            cursor.movePosition( QTextCursor::StartOfBlock );
            cursor.movePosition( QTextCursor::Right, QTextCursor::MoveAnchor, position );
            cursor.movePosition( QTextCursor::Right, QTextCursor::KeepAnchor, length );
            cursor.insertImage( _resourceName( iter.value() ) );

            // update text and position
            text = cursor.block().text();
            position++;

        } else position += length;
    }
}

//________________________________________________________________________
QString EmoticonProvider::revert( QTextCursor cursor, bool& changed ) const
{
    Debug::Throw( "EmoticonProvider::revert.\n" );

    // reset changed flag
    changed = false;

    // get text from cursor
    const QString text( cursor.selection().toPlainText() );
    QString copy( text );

    // get cursor position
    const int cursorPosition( qMin( cursor.position(), cursor.anchor() ) );
    int position(0);
    int offset(0);
    while( (position = text.indexOf( QChar( 0xFFFC ), position ) ) >= 0 )
    {
        cursor.setPosition( cursorPosition );
        cursor.movePosition( QTextCursor::Right, QTextCursor::MoveAnchor, position+1 );
        QTextCharFormat format( cursor.charFormat() );
        if( !format.isImageFormat() )
        {
            ++position;
            continue;
        }

        // find matching emoticon
        const QString resourceName( format.toImageFormat().name() );
        foreach( const auto& emoticon, emoticons_ )
        {
            if( _resourceName( emoticon ) != resourceName ) continue;

            // replace in copy
            copy.replace( position+offset, 1, emoticon.strings().front() );
            offset += emoticon.strings().front().size() - 1;

            // update changed flag
            changed = true;
            break;
        }

        ++position;
    }

    return copy;

}

//________________________________________________________________________
QString EmoticonProvider::_resourceName( const Emoticon& emoticon ) const
{ return QString( "images%1" ).arg( emoticon.resourceName() ); }
