/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "IdleTime.h"
#include "Debug.h"
#include "XcbUtil.h"

#include <QThread>

#if HAVE_XCB
#include <xcb/xcb.h>
#include <xcb/screensaver.h>
#include <xcb/record.h>
#endif

class ActivityThread: public QThread
{

    public:

    //* constructor
    ActivityThread( QObject* = nullptr );

    //* destructor
    virtual ~ActivityThread( void ) = default;

    protected:

    //* Check files validity. Post a ValidFileEvent when finished
    virtual void run( void );

};

//_________________________________________________________
ActivityThread::ActivityThread( QObject* parent ):
    QThread( parent )
{}

//_________________________________________________________
void ActivityThread::run( void )
{
    #if HAVE_XCB

    // connection
    xcb_connection_t* connection = xcb_connect( nullptr, nullptr );

    // record range
    xcb_record_range_t range;
    memset(&range, 0, sizeof(range));
    range.device_events.first = XCB_KEY_PRESS;
    range.device_events.last = XCB_MOTION_NOTIFY;

    // client spec
    const xcb_record_client_spec_t clientSpec( XCB_RECORD_CS_ALL_CLIENTS );

    // create record context
    xcb_record_context_t contect = xcb_generate_id( connection );
    {
        xcb_void_cookie_t cookie( xcb_record_create_context_checked( connection, contect, 0, 1, 1, &clientSpec, &range ) );
        xcb_generic_error_t* error( xcb_request_check( connection, cookie ) );
        if (error)
        {
            free(error);
            xcb_flush( connection );
            xcb_disconnect( connection );
            return;
        }
    }

    // enable recording
    xcb_record_enable_context_cookie_t cookie( xcb_record_enable_context( connection, contect ) );

    bool done = false;
    while( !done )
    {
        // get context reply
        xcb_record_enable_context_reply_t *reply = xcb_record_enable_context_reply( connection, cookie, nullptr );
        if( !reply ) break;

        // check category 0 means recordFromServer
        if( reply->category == 0 ) done = true;
        free( reply );
    }

    // cleanup
    xcb_record_disable_context( connection, contect );
    xcb_record_free_context( connection, contect );
    xcb_flush( connection );
    xcb_disconnect( connection );

    #endif

}

//_________________________________________________________
IdleTime::IdleTime( QObject* parent ):
    QObject( parent ),
    Counter( "IdleTime" )
{
    Debug::Throw( "IdleTime::IdleTime.\n" );
    activityThread_ = new ActivityThread( this );
    connect( activityThread_, SIGNAL(finished()), this, SLOT(_activityDetected()) );
}

//_________________________________________________________
bool IdleTime::isSupported( void ) const
{
    Debug::Throw( "IdleTime::isSupported.\n" );
    #if HAVE_XCB
    return true;
    #else
    return false;
    #endif
}

//_________________________________________________________
bool IdleTime::isRunning( void ) const
{ return timer_.isActive() || activityThread_->isRunning(); }

//_________________________________________________________
void IdleTime::setTimeout( int value )
{
    Debug::Throw( "IdleTime::setTimeout.\n" );

    // do nothing if value is unchanged
    if( timeout_ == value ) return;

    // assign
    timeout_ = value;

    // if running, restart, with new timeout
    if( timer_.isActive() )
    {
        timer_.stop();
        _checkTimeout();
    }

    return;

}

//_________________________________________________________
void IdleTime::start( void )
{
    Debug::Throw( "IdleTime::start.\n" );
    if( !isSupported() ) Debug::Throw(0, "IdleTime::start - warning: not supported.\n" );
    if( isRunning() ) Debug::Throw(0, "IdleTime::start - warning: already running.\n" );
    else {

        isActive_ = true;
        _checkTimeout();

    }

}

//_________________________________________________________
void IdleTime::stop( void )
{
    Debug::Throw( "IdleTime::stop.\n" );
    if( timer_.isActive() ) timer_.stop();
    isActive_ = false;
}

//_________________________________________________________
void IdleTime::timerEvent( QTimerEvent* event )
{
    Debug::Throw( "IdleTime::timerEvent.\n" );
    if( event->timerId() == timer_.timerId() )
    {
        timer_.stop();
        _checkTimeout();
    } else return QObject::timerEvent( event );
}

//_________________________________________________________
void IdleTime::_activityDetected( void )
{
    Debug::Throw( "IdleTime::_activityDetected.\n" );
    emit idleStateChanged( false );
    if( isSupported() && isActive_ && timeout_ >= 0 ) timer_.start( timeout_, this );
}

//_________________________________________________________
void IdleTime::_checkTimeout( void )
{

    Debug::Throw( "IdleTime::_checkTimeout.\n" );

    // do nothing if no timeout
    if( !(isSupported() && timeout_ > 0) ) return;

    // get current idle time
    uint32_t idleTime( _idleTime() );
    if( idleTime >= (uint32_t)timeout_ )
    {

        emit idleStateChanged( true );

        // start thread
        if( !activityThread_->isRunning() ) activityThread_->start();

    } else {

        timer_.start( timeout_ - idleTime, this );

    }

}

//_________________________________________________________
uint32_t IdleTime::_idleTime( void ) const
{
    Debug::Throw( "IdleTime::_idleTime.\n" );
    #if HAVE_XCB
    xcb_connection_t* connection( XcbUtil::get().connection<xcb_connection_t>() );
    xcb_screensaver_query_info_cookie_t cookie( xcb_screensaver_query_info( connection, XcbUtil::get().appRootWindow() ) );
    xcb_screensaver_query_info_reply_t* info( xcb_screensaver_query_info_reply( connection, cookie, nullptr ) );

    const uint32_t idle = info->ms_since_user_input;
    free (info);
    return idle;

    #else
    return 0;
    #endif
}
