/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "MainWindow.h"

#include "ClientApplication.h"
#include "Command.h"
#include "ConnectedClientsList.h"
#include "CustomToolBar.h"
#include "CustomToolButton.h"
#include "EmoticonProvider.h"
#include "EmoticonSelectionMenu.h"
#include "HistoryMenu.h"
#include "HtmlHighlighter.h"
#include "IconNames.h"
#include "IconEngine.h"
#include "IdleTime.h"
#include "InformationDialog.h"
#include "LineEditor.h"
#include "LogWidget.h"
#include "Menu.h"
#include "MessageBoardWidget.h"
#include "MessageEditorWidget.h"
#include "OpenLinkDialog.h"
#include "ProgressStatusBar.h"
#include "QuestionDialog.h"
#include "Singleton.h"
#include "TimeStamp.h"
#include "TreeView.h"

#include <QLayout>
#include <QSystemTrayIcon>

//_________________________________________________
MainWindow::MainWindow( QWidget* parent ):
    BaseMainWindow( parent ),
    Counter( "MainWindow" ),
    connectedClientsModel_( this )
{

    Debug::Throw( "MainWindow::MainWindow.\n" );
    setOptionName( "MAIN_WINDOW" );

    // actions
    _installActions();
    ClientApplication& application( *Singleton::get().application<ClientApplication>() );
    addAction( &application.closeAction() );

    // main widget and layout
    QWidget* main( new QWidget( this ) );
    setCentralWidget( main );

    QHBoxLayout* hLayout = new QHBoxLayout();
    hLayout->setSpacing(2);
    hLayout->setMargin(0);
    main->setLayout(hLayout);

    // connected clients list
    connectedClientsList_ = new ConnectedClientsList( main );
    connectedClientsList_->setProperty( "_kde_side_panel_view", false );
    connectedClientsList_->setModel( &connectedClientsModel_ );
    hLayout->addWidget( connectedClientsList_ );

    // message board
    QVBoxLayout* vLayout = new QVBoxLayout();
    vLayout->setMargin(0);
    vLayout->setSpacing(2);
    hLayout->addLayout( vLayout );

    messageBoardWidget_ = new MessageBoardWidget();
    messageBoardWidget_->setReadOnly( true );
    messageBoardWidget_->viewport()->installEventFilter( this );

    TextEditor::Container* container = new TextEditor::Container( main, messageBoardWidget_ );
    vLayout->addWidget( container );

    // horizontal layout
    hLayout = new QHBoxLayout();
    hLayout->setMargin(0);
    vLayout->addLayout( hLayout );

    // message editor
    messageEditorWidget_ = new MessageEditorWidget( main );
    messageEditorWidget_->installEventFilter( this );
    connect( messageEditorWidget_, SIGNAL(textEdited(QString)), this, SLOT(_editorModified()) );
    hLayout->addWidget( messageEditorWidget_ );

    // emoticons
    QToolButton* emoticonButton = new QToolButton( main );
    {
        emoticonButton->setIcon( QIcon( ":/smile.svgz" ) );
        emoticonButton->setAutoRaise( true );
        EmoticonSelectionMenu* menu = new EmoticonSelectionMenu();
        emoticonButton->setMenu( menu );
        emoticonButton->setPopupMode( QToolButton::InstantPopup );
        emoticonButton->setIconSize( IconSize( IconSize::Small ) );
        connect( menu, SIGNAL(emoticonSelected(Emoticon)), this, SLOT(_insertEmoticon(Emoticon)) );
    }

    hLayout->addWidget( emoticonButton );

    // statusBar
    setStatusBar( statusBar_ = new ProgressStatusBar( this ) );
    statusBar_->addClock();

    // menu bar
    Menu* menu( new Menu( this ) );
    setMenuBar( menu );

    // toolbar
    CustomToolBar* toolbar;
    toolbar = new CustomToolBar( tr( "Main Toolbar" ), this, "MAIN_TOOLBAR" );
    toolbar->addAction( toggleConnectionAction_ );
    toolbar->addWidget( userStateButton_ );
    toolbar->addWidget( historyButton_ );

    // reset replace iterator
    _clearReplaceIterator();

    // connections
    messageEditorWidget_->setFocus();
    connect( messageEditorWidget_, SIGNAL(returnPressed()), this, SLOT(_sendMessage()) );
    connect( &messageEditorWidget_->cancelMessageEditionAction(), SIGNAL(triggered()), this, SLOT( _cancelMessageEdition()) );
    connect( &messageEditorWidget_->removeMessageAction(), SIGNAL(triggered()), this, SLOT(_removeReplaceMessage()) );
    connect( &messageBoardWidget_->editMessageAction(), SIGNAL(triggered()), this, SLOT(_editContextMenuMessage()) );
    connect( &messageBoardWidget_->removeMessageAction(), SIGNAL(triggered()), this, SLOT(_removeContextMenuMessage()) );
    connect( &messageBoardWidget_->openLinkAction(), SIGNAL(triggered()), this, SLOT(_openLink()) );
    connect( messageBoardWidget_, SIGNAL(linkActivated(QString)),  this, SLOT(_openLink(QString)) );

    // disable everything
    updateConnectionState( id_, ClientInterface::Disconnected );

    // configuration
    connect( Singleton::get().application(), SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    _updateConfiguration();

}

//_________________________________________________
void MainWindow::setId( ConnectionId id )
{

    Debug::Throw() << "MainWindow::setId - id: " << id << endl;

    // remove old id from model and update
    connectedClientsModel_.remove( id_ );
    id_ = id;

    // re-add to model with correct status
    if( state_ == ClientInterface::Disconnected ) id.setUserState( ConnectionId::Offline );
    connectedClientsModel_.add( id );

    // update window title
    if( !id_.user().isEmpty() ) setWindowTitle( QString( "%1 - Homechat" ).arg( id_.userString() ) );
    else setWindowTitle( "Homechat" );

    // update user state button icon
    _updateUserStateButton();
}

//_________________________________________________
MainWindow::~MainWindow( void ) = default;

//_________________________________________________
bool MainWindow::eventFilter( QObject* object, QEvent* event )
{
    if( object == messageEditorWidget_ )
    {

        switch( event->type() )
        {

            case QEvent::KeyPress:
            {
                QKeyEvent *keyEvent = static_cast<QKeyEvent*>(event);
                switch( keyEvent->key() )
                {
                    default: break;
                    case Qt::Key_Up:
                    {

                        // check message list
                        if( messageList_.isEmpty() ) break;

                        // find previous message in stack that matches same user
                        ChatMessage::List::iterator iterator( replaceIterator_ );
                        while( iterator != messageList_.begin() )
                        {
                            --iterator;
                            if( ( iterator->id() == id_ && !iterator->text().isEmpty() ) || iterator == messageList_.begin() ) break;
                        }

                        // if found, update editor text
                        if( iterator->id() == id_ )
                        {
                            replaceIterator_ = iterator;
                            replaceMessage_ = *iterator;
                            messageEditorWidget_->setText( replaceMessage_.text() );
                        }

                        break;
                    }

                    case Qt::Key_Down:
                    {

                        // check message list and iterator
                        if( messageList_.isEmpty() ) break;
                        if( replaceIterator_ == messageList_.end() ) break;

                        // find previous message in stack that matches same user
                        ChatMessage::List::iterator iterator( replaceIterator_ );
                        while( iterator != messageList_.end() )
                        {
                            ++iterator;
                            if( iterator == messageList_.end() || ( iterator->id() == id_ && !iterator->text().isEmpty() ) ) break;
                        }

                        // if found, update editor text
                        if( iterator == messageList_.end() )
                        {

                            messageEditorWidget_->clear();
                            _clearReplaceIterator();

                        } else {

                            replaceIterator_ = iterator;
                            replaceMessage_ = *iterator;
                            messageEditorWidget_->setText( replaceMessage_.text() );

                        }

                        break;
                    }

                }

            }

            case QEvent::ContextMenu:
            {
                messageEditorWidget_->cancelMessageEditionAction().setEnabled( replaceIterator_ != messageList_.end() );
                messageEditorWidget_->removeMessageAction().setEnabled( replaceIterator_ != messageList_.end() );
                break;
            }

            default: break;

        }

    } else if( object == messageBoardWidget_->viewport() && event->type() == QEvent::ContextMenu ) {

        // find text cursor and block number
        const QTextCursor textCursor( messageBoardWidget_->cursorForPosition( static_cast<QContextMenuEvent*>( event )->pos() ) );
        const int blockNumber( textCursor.blockNumber() );

        // find matching message
        contextMenuMessage_ = ChatMessage();
        foreach( const auto& message, messageSet_ )
        {
            const auto& blockRange( message.blockRange() );
            if( blockNumber >= blockRange.first && blockNumber <= blockRange.second )
            {
                contextMenuMessage_ = message;
                break;
            }
        }

        messageBoardWidget_->editMessageAction().setEnabled( contextMenuMessage_.isValid() && contextMenuMessage_.id() == id_ );
        messageBoardWidget_->removeMessageAction().setEnabled( contextMenuMessage_.isValid() && contextMenuMessage_.id() == id_ );

    }

    // base class
    return BaseMainWindow::eventFilter( object, event );

}

//_________________________________________________
void MainWindow::updateConnectionState( ConnectionId id, ClientInterface::State state )
{
    Debug::Throw()
        << "MainWindow::updateConnectionState -"
        << " user: " << id.user()
        << " state: " << state
        << " connection time: " << id.connectionTime()
        << " disconnection time: " << id.disconnectionTime()
        << endl;

    // update connected clients list
    switch( state )
    {
        case ClientInterface::AwaitingReply:
        case ClientInterface::Connected:
        {
            // add to model
            connectedClientsModel_.add( id );

            // update ping clients action
            pingClientsAction_->setEnabled( _connectedClientsCount() > 1 );

            if( id != id_ )
            { emit newNotification( QString( tr( "%1 has joined" ) ).arg( id.userString() ) ); }

            break;
        }

        default:
        case ClientInterface::Disconnected:
        {

            // update model
            if( connectedClientsModel_.index( id ).isValid() )
            {
                id.setUserState( ConnectionId::Offline );
                id.setDisconnectionTime( TimeStamp::now() );
                connectedClientsModel_.add( id );
            }

            // update ping clients action
            pingClientsAction_->setEnabled( _connectedClientsCount() > 1 );

            // emit notification
            if( id != id_)
            { emit newNotification( QString( tr( "%1 has quit" ) ).arg( id.userString() ) ); }
            break;
        }

    }

    if( id == id_ )
    {

        // update internal state
        state_ = state;

        // copy id, to get right connection time
        id_ = id;

        // update local widgets, actions etc.
        QList<QWidget*> widgets = QList<QWidget*>() << historyButton_;

        switch( state )
        {
            case ClientInterface::AwaitingReply:
            {
                // update widget enable state
                foreach( auto widget, widgets ) { widget->setEnabled( false ); }
                userStateActions_->setEnabled( false );

                // update toggle connection action
                toggleConnectionAction_->setEnabled( true );
                toggleConnectionAction_->setIcon( IconEngine::get( IconNames::Connect ) );
                toggleConnectionAction_->setText( tr( "Sign Out" ) );

                // start idle time monitoring
                if( idleTime_->isSupported() && !idleTime_->isRunning() ) idleTime_->start();

                break;
            }

            case ClientInterface::Connected:
            {
                // update widget enable state
                foreach( auto widget, widgets ) { widget->setEnabled( true ); }
                userStateActions_->setEnabled( true );

                // update toggle connection action
                toggleConnectionAction_->setEnabled( true );
                toggleConnectionAction_->setIcon( IconEngine::get( IconNames::Connect ) );
                toggleConnectionAction_->setText( tr( "Sign Out" ) );

                // start idle time monitoring
                if( idleTime_->isSupported() && !idleTime_->isRunning() ) idleTime_->start();

                break;
            }

            default:
            case ClientInterface::Disconnected:
            {

                // update widget enable state
                foreach( auto widget, widgets ) { widget->setEnabled( false ); }
                userStateActions_->setEnabled( false );

                // update toggle connection action
                toggleConnectionAction_->setEnabled( true );
                toggleConnectionAction_->setIcon( IconEngine::get( IconNames::Disconnect ) );
                toggleConnectionAction_->setText( tr( "Sign In" ) );

                // stop idle time monitoring
                idleTime_->stop();

                break;
            }

        }

        _updateUserStateButton();
        _updateStatusBar();

    }

}

//_________________________________________________
void MainWindow::updateIdleState( bool isIdle )
{
    Debug::Throw() << "MainWindow::updateIdleState - idle: " << isIdle << endl;
    if( state_ != ClientInterface::Connected ) return;
    if( isIdle )
    {

        if( id_.userState() == ConnectionId::Online )
        { updateUserState( ConnectionId::Away ); }

    } else if( id_.userState() == ConnectionId::Away ) {

        updateUserState( ConnectionId::Online );

    }
}

//_________________________________________________
void MainWindow::updateUserState( ConnectionId::UserState state )
{
    Debug::Throw( "MainWindow::updateUserState.\n" );

    // update id
    id_.setUserState( state );

    // update connectedClientsModel
    if( state_ == ClientInterface::Connected )
    {
        connectedClientsModel_.add( id_ );
        emit userStateChanged( state );
    }

    _updateUserStateButton();
    _updateStatusBar();

}

//_________________________________________________
void MainWindow::updateUserState( ConnectionId id )
{
    Debug::Throw( "MainWindow::updateUserState.\n" );

    // do nothing if id is invalid or equal to  this one
    if( !id.isValid() || id == id_ ) return;

    // get old state
    ConnectionId::UserState oldState( ConnectionId::Offline );
    const QModelIndex index( connectedClientsModel_.index( id ) );
    if( index.isValid() ) oldState = connectedClientsModel_.get( index ).userState();

    // update model
    if( id.userState() == ConnectionId::Renamed ) connectedClientsModel_.remove( id );
    else if( id.userState() != ConnectionId::Offline || index.isValid() )
    { connectedClientsModel_.add( id ); }

    // emit notification, unless previously offline
    if( oldState != ConnectionId::Offline && oldState != id.userState() )
    {
        QString message;
        switch( id.userState() )
        {
            default: break;
            case ConnectionId::Online: message = QString( tr( "%1 is online" ) ).arg( id.userString() ); break;
            case ConnectionId::Away:
            case ConnectionId::AwayForced: message = QString( tr( "%1 is away" ) ).arg( id.userString() ); break;
            case ConnectionId::Unavailable: message = QString( tr( "%1 is unavailable" ) ).arg( id.userString() ); break;
        }

        if( !message.isNull() )
        { emit newNotification( message ); }
    }

}

//_________________________________________________
void MainWindow::updateBusyState( bool busy, int size )
{
    Debug::Throw() << "MainWindow::updateBusyState - busy: " << busy << endl;
    busy_ = busy;
    if( busy_ )
    {
        // need at lest 10 messages to trigger progressbar
        if( showFetchProgress_ && size > 10 )
        {
            statusBar_->showProgressBar();
            statusBar_->progressBar().setMaximum( size );
            statusBar_->progressBar().reset();
        }

    } else statusBar_->showLabel();
}

//_________________________________________________
void MainWindow::printError( QString error, bool tryReconnect )
{
    Debug::Throw( "MainWindow::printError.\n" );
    if( isHidden() ) emit newNotification( error );
    else if( tryReconnect ) {

        // dialog must be modal to main window only
        // it should not prevent systray to work, for instance
        error += tr( "\nTry reconnect ?" );
        QuestionDialog* dialog = new QuestionDialog( this, error );
        dialog->setAttribute( Qt::WA_DeleteOnClose, true );
        dialog->setWindowModality( Qt::WindowModal );
        dialog->okButton().setText( tr( "Reconnect" ) );
        dialog->okButton().setIcon( IconEngine::get( IconNames::Connect ) );
        connect( &dialog->okButton(), SIGNAL(clicked()), toggleConnectionAction_, SIGNAL(triggered()) );
        dialog->show();

    } else {

        // dialog must be modal to main window only
        // it should not prevent systray to work, for instance
        InformationDialog* dialog = new InformationDialog( this, error );
        dialog->setAttribute( Qt::WA_DeleteOnClose, true );
        dialog->setWindowModality( Qt::WindowModal );
        dialog->show();

    }

}

//_________________________________________________
void MainWindow::printMessage( ChatMessage message )
{
    Debug::Throw() << "MainWindow::printMessage - user: " << message.id().user() << " text: " << message.text() << endl;

    // check text and id
    if( !message.isValid() ) return;
    if( !message.receptionTime().isValid() ) message.setReceptionTime( TimeStamp::now() );

    // increment progressbar if visible
    if( showFetchProgress_ && statusBar_->progressBar().isVisible() )
    { statusBar_->progressBar().setValue( statusBar_->progressBar().value()+1 ); }

    // append message
    if( _appendMessage( message ) )
    {

        // clear message board and last printed message
        messageBoardWidget_->clear();
        lastMessage_ = ChatMessage();

        // print full list
        _printMessageList();

    } else _printMessage( message );

    if( !( busy_ || message.id() == id_ ) )
    {

        // emit notification
        if( message.text().isEmpty() ) emit newNotification( QString( tr( "%1 has removed a message" ) ).arg( message.id().userString() ) );
        else {

            emit newNotification( QString( tr( "%1 says: '%2'" ) )
                .arg( message.id().userString() )
                .arg( message.text() ) );

            // emit pending messages
            if( state_ == ClientInterface::Connected && !isActiveWindow() ) emit hasPendingMessages(true);

        }

    }

}

//____________________________________
void MainWindow::changeEvent( QEvent* event )
{
    switch( event->type() )
    {
        default: break;
        case QEvent::ActivationChange:
        {
            if( state_ == ClientInterface::Connected && isActiveWindow() )
            { emit hasPendingMessages( false ); }
            break;
        }
    }

    // base class method
    return BaseMainWindow::changeEvent( event );

}

//____________________________________
void MainWindow::closeEvent( QCloseEvent *event )
{
    Debug::Throw( "MainWindow::closeEvent.\n" );
    event->accept();

    // trigger close action, unless systemtray icon is used
    if( !( QSystemTrayIcon::isSystemTrayAvailable() && XmlOptions::get().get<bool>( "USE_SYSTEMTRAY" ) ) )
    {

        Singleton::get().application<ClientApplication>()->closeAction().trigger();

    } else if( !XmlOptions::get().contains( "SHOW_CLOSE_DIALOG" ) || XmlOptions::get().get<bool>( "SHOW_CLOSE_DIALOG" ) ) {

        SingleTimeDialog dialog( this, tr(
            "Closing the main window will keep Homechat running in the System Tray. "
            "Use <b>Exit</b> from the menu, or the Homechat tray icon to exit the application." ) );

        dialog.setWindowTitle( tr( "System Tray - Homechat" ) );
        dialog.exec();
        XmlOptions::get().set<bool>( "SHOW_CLOSE_DIALOG", !dialog.dontShowAgain() );

    }

}

//_______________________________________________________
void MainWindow::timerEvent( QTimerEvent* event )
{

    if( event->timerId() == userTypingTimer_.timerId() )
    {

        // stop timer and save time
        userTypingTimer_.stop();
        _updateStatusBar();

    } else if( event->timerId() == thisUserTypingTimer_.timerId() ) {

        thisUserTypingTimer_.stop();

    } else return BaseMainWindow::timerEvent( event );

}

//_________________________________________________
void MainWindow::clearConnectedClients( void )
{
    Debug::Throw( "MainWindow::clearConnectedClients.\n" );
    connectedClientsModel_.clear();
}

//_________________________________________________
void MainWindow::clearMessageList( void )
{
    Debug::Throw( "MainWindow::clearMessageList.\n" );

    // clear messages
    messageSet_.clear();
    messageList_.clear();
    _clearReplaceIterator( false );

    // clear board
    lastMessage_ = ChatMessage();
    messageBoardWidget_->clear();
}

//_________________________________________________
void MainWindow::notifyUserTyping( ConnectionId id )
{
    Debug::Throw( "MainWindow::notifyUserTyping.\n" );
    statusBar_->label().setTextAndUpdate( QString( tr( "%1 is writing a message ..." ) ).arg( id.userString() ) );
    if( userTypingTimer_.isActive() ) userTypingTimer_.stop();
    userTypingTimer_.start( userTypingTimeout_, this );
}

//_________________________________________________
void MainWindow::_toggleConnection( void )
{
    Debug::Throw( "MainWindow::_toggleConnection.\n" );
    switch( state_ )
    {
        default:
        case ClientInterface::Disconnected:
        emit requestConnection();
        break;

        case ClientInterface::AwaitingReply:
        case ClientInterface::Connected:
        emit requestDisconnection();
        break;
    }

    return;

}

    //_________________________________________________
void MainWindow::_sendMessage( void )
{
    const QString text( messageEditorWidget_->text() );
    if( text.isEmpty() )
    {
        // reset replace iterator and do nothing
        _clearReplaceIterator();
        return;
    }

    // create message and print
    ChatMessage message( id_, text );

    // adjust creation and reception time from replaceIterator if valid
    if( replaceMessage_.isValid() )
    {
        message.setCreationTime( replaceMessage_.creationTime() );
        message.setReceptionTime( replaceMessage_.receptionTime() );
        _clearReplaceIterator();
    }

    // print
    printMessage( message );

    // send to server
    emit newMessage( message );
    messageEditorWidget_->clear();

}

//________________________________________________________
void MainWindow::_userStateActionTriggered( QAction* action )
{
    Debug::Throw( "MainWindow::_userStateActionTriggered.\n" );

    ActionMap::const_iterator iter( userStateActionMap_.find( action ) );
    if( iter != userStateActionMap_.constEnd() )
    { updateUserState( iter.value() ); }

}

//________________________________________________________
void MainWindow::_insertEmoticon( Emoticon emoticon )
{ messageEditorWidget_->insert( emoticon.strings().front() ); }

//________________________________________________________
void MainWindow::_editorModified( void )
{
    // check timer
    if( thisUserTypingTimer_.isActive() ) return;

    /*
    emit signal and start timer
    this prevents too many signals to be sent
    */
    thisUserTypingTimer_.start( userTypingTimeout_/2, this );
    emit userTyping();
}

//______________________________________
void MainWindow::_editContextMenuMessage( void )
{
    Debug::Throw( "MainWindow::_editContextMenuMessage.\n" );
    if( !( contextMenuMessage_.isValid() && contextMenuMessage_.id() == id_ ) ) return;

    // copy context menu message into replaceMessage_, and update iterator
    replaceMessage_ = contextMenuMessage_;
    _updateReplaceIterator();

    // update editor
    messageEditorWidget_->setText( contextMenuMessage_.text() );

}

//______________________________________
void MainWindow::_removeContextMenuMessage( void )
{

    if( !( contextMenuMessage_.isValid() && contextMenuMessage_.id() == id_ ) ) return;

    Debug::Throw()
        << "MainWindow::_editContextMenuMessage - message: "
        << contextMenuMessage_
        << endl;

    contextMenuMessage_.setText( QString() );

    // print
    printMessage( contextMenuMessage_ );

    // send to server
    emit newMessage( contextMenuMessage_ );

}

//______________________________________
void MainWindow::_removeReplaceMessage( void )
{

    if( replaceIterator_ == messageList_.end() ) return;

    Debug::Throw()
        << "MainWindow::_removeReplaceMessage - message: "
        << *replaceIterator_
        << endl;

    replaceIterator_->setText( QString() );
    replaceMessage_ = *replaceIterator_;

    // print
    printMessage( replaceMessage_ );

    // send to server
    emit newMessage( replaceMessage_ );

    // reset replace iterator
    _clearReplaceIterator();

    // clear editor widget
    messageEditorWidget_->clear();

}

//______________________________________
void MainWindow::_cancelMessageEdition( void )
{
    Debug::Throw( "MainWindow::_cancelMessageEdition.\n" );
    if( replaceIterator_ == messageList_.end() ) return;

    // clear replace iterator
    _clearReplaceIterator();

    // clear message editor
    messageEditorWidget_->clear();
}

//_____________________________________________
void MainWindow::_openLink( void )
{
    Debug::Throw( "MainWindow::_openLink.\n" );
    QString anchor( messageBoardWidget_->anchor() );
    if( anchor.isEmpty() ) return;

    OpenLinkDialog dialog( this, anchor );
    dialog.setWindowTitle( tr( "Open Link - Homechat" ) );

    // retrieve applications from options
    Options::List applications( XmlOptions::get().specialOptions( "OPEN_LINK_APPLICATIONS" ) );
    foreach( const Option& option, applications )
    { dialog.actionComboBox().addItem( option.raw() ); }

    if( dialog.centerOnParent().exec() == QDialog::Rejected ) return;

    // retrieve application from combobox and add as options
    QString command( dialog.actionComboBox().currentText() );
    if( command.isEmpty() )
    {
        InformationDialog( this, tr( "No command specified to open the selected files. <Open> canceled." ) ).exec();
        return;
    }

    // update options
    XmlOptions::get().add( "OPEN_LINK_APPLICATIONS", Option( command, Option::Recordable|Option::Current ) );

    // execute
    ( Command( command ) << anchor ).run();

}

//_____________________________________________
void MainWindow::_openLink( QString anchor )
{

    Debug::Throw( "MainWindow::_openLink.\n" );
    if( anchor.isEmpty() ) return;
    const QString command( XmlOptions::get().raw( "XDG_OPEN" ) );
    if( File( command ).exists() )
    { ( Command( command ) << anchor ).run(); }

}

//______________________________________
void MainWindow::_updateConfiguration( void )
{
    Debug::Throw( "MainWindow::_updateConfiguration.\n" );
    idleTime_->setTimeout( 1000 * 60 * XmlOptions::get().get<int>( "IDLE_TIMEOUT" ) );
    showFetchProgress_ = XmlOptions::get().get<bool>( "SHOW_FETCH_PROGRESS" );

    bool changed( false );
    if( messageBoardWidget_->htmlHighlighter().setEnabled( XmlOptions::get().get<bool>( "SHOW_LINKS" ) ) ) changed = true;
    if( messageBoardWidget_->emoticonProvider().setEnabled( XmlOptions::get().get<bool>( "USE_EMOTICONS" ) ) ) changed = true;
    if( messageBoardWidget_->emoticonProvider().setIconSize( XmlOptions::get().get<int>( "EMOTICON_SIZE" ) ) ) changed = true;
    if( changed  )
    {

       // clear message board and last printed message
        messageBoardWidget_->clear();
        lastMessage_ = ChatMessage();

        // print full list
        _printMessageList();

    }

    return;
}

//_________________________________________________
void MainWindow::_installActions( void )
{
    Debug::Throw( "MainWindow::_installActions.\n" );

    // toggle connection
    addAction( toggleConnectionAction_ = new QAction( IconEngine::get( IconNames::Connect ), tr( "Connect" ), this ) );
    connect( toggleConnectionAction_, SIGNAL(triggered()), this, SLOT(_toggleConnection()) );

    // connected clients
    addAction( fetchConnectedClientsAction_ = new QAction( IconEngine::get( IconNames::Find ), tr( "Fetch connected clients" ), this ) );

    // server ping
    addAction( pingServerAction_ = new QAction( tr( "Check Connection to Server" ), this ) );

    //  client ping
    addAction( pingClientsAction_ = new QAction( tr( "Check Users Connection" ), this ) );

    // reload
    addAction( reloadAction_ = new QAction( IconEngine::get( IconNames::Reload ), tr( "Reload Messages" ), this ) );
    reloadAction_->setShortcut( QKeySequence::Refresh );

    // idle time
    idleTime_ = new IdleTime( this );
    connect( idleTime_, SIGNAL(idleStateChanged(bool)), this, SLOT(updateIdleState(bool)) );

    // history
    historyButton_ = new CustomToolButton();
    historyButton_->setIcon( IconEngine::get( IconNames::History ) );
    historyButton_->setText( tr( "Messages History" ) );
    historyButton_->setToolTip( tr( "Show messages history" ) );
    historyButton_->setMenu( historyMenu_ = new HistoryMenu( this ) );
    historyButton_->setPopupMode( QToolButton::InstantPopup );

    // user state actions
    using ActionPair = QPair<QAction*, ConnectionId::UserState>;
    QVector<ActionPair> actions;
    actions
        << ActionPair( new QAction( IconEngine::get( IconNames::UserOnline ), tr( "Online" ), this ), ConnectionId::Online )
        << ActionPair( new QAction( IconEngine::get( IconNames::UserAway ), tr( "Away" ), this ), ConnectionId::AwayForced )
        << ActionPair( new QAction( IconEngine::get( IconNames::UserBusy ), tr( "Unavailable" ), this ), ConnectionId::Unavailable );

    userStateActions_ = new QActionGroup( this );
    foreach( const auto& actionPair, actions )
    {
        userStateActionMap_.insert( actionPair.first, actionPair.second );
        userStateActions_->addAction( actionPair.first );
    }

    connect( userStateActions_, SIGNAL(triggered(QAction*)), SLOT(_userStateActionTriggered(QAction*)) );

    // user state button
    userStateButton_ = new CustomToolButton();
    userStateButton_->setIcon( IconEngine::get( IconNames::UserOnline ) );
    userStateButton_->setText( tr( "Status" ) );
    userStateButton_->setToolTip( tr( "Change status" ) );
    userStateButton_->setMenu( new QMenu() );
    userStateButton_->setPopupMode( QToolButton::InstantPopup );
    userStateButton_->menu()->addActions( userStateActions_->actions() );

}

//_________________________________________________
bool MainWindow::_appendMessage( const ChatMessage& message )
{
    Debug::Throw( "MainWindow::_appendMessage.\n" );

    // first check if message is in set
    ChatMessage::Set::iterator iter( messageSet_.find( message ) );
    if( iter != messageSet_.end() )
    {

        Debug::Throw( "MainWindow::_appendMessage - found existing message.\n" );

        // erase old message
        messageSet_.erase( iter );

        // insert new message if text is not empty
        if( !message.text().isEmpty() )
        { messageSet_.insert( message ); }

        // regenerate list
        messageList_ = messageSet_.toList();
        qSort( messageList_.begin(), messageList_.end(), ChatMessage::SortReceptionTimeFTor() );

        // restore replace iterator
        _updateReplaceIterator();

        return true;

    } else if( message.text().isEmpty() ) {

        // do nothing and return false if message is empty
        return false;

    } else {

        // insert in set
        messageSet_.insert( message );

        // try append at the back of the list
        if( messageList_.isEmpty() || messageList_.last().receptionTime() <= message.receptionTime() )
        {

            messageList_.append( message );
            _updateReplaceIterator();
            return false;

        } else {

            // regenerate sorted list from set
            messageList_ = messageSet_.toList();
            qSort( messageList_.begin(), messageList_.end(), ChatMessage::SortReceptionTimeFTor() );
            _updateReplaceIterator();
            return true;

        }

    }

}

//_________________________________________________
void MainWindow::_clearReplaceIterator( bool clearMessage )
{
    replaceIterator_ = messageList_.end();
    if( clearMessage ) replaceMessage_ = ChatMessage();
}

//_________________________________________________
void MainWindow::_updateStatusBar( void ) const
{

    QString message;
    switch( state_ )
    {
        case ClientInterface::Disconnected:
        message = tr( "Offline" );
        break;

        case ClientInterface::AwaitingReply:
        message = tr( "Waiting for server authorization" );
        break;

        default:
        case ClientInterface::Connected:
        {
            switch( id_.userState() )
            {
                default:
                case ConnectionId::Online: message = tr( "Online" ); break;
                case ConnectionId::Away:
                case ConnectionId::AwayForced: message = tr( "Away" ); break;
                case ConnectionId::Unavailable: message = tr( "Unavailable" ); break;
                case ConnectionId::Offline: message = tr( "Offline" ); break;
            }
            break;
        }

    }

    // update
    statusBar_->label().setTextAndUpdate( message );

}

//_________________________________________________
void MainWindow::_updateUserStateButton( void ) const
{ userStateButton_->setEnabled( state_ == ClientInterface::Connected ); }

//_________________________________________________
void MainWindow::_updateReplaceIterator( void )
{
    if( !replaceMessage_.isValid() ) replaceIterator_ = messageList_.end();
    else replaceIterator_ = qFind( messageList_.begin(), messageList_.end(), replaceMessage_ );
}

//_________________________________________________
void MainWindow::_printMessageList( void )
{ foreach( const auto& message, messageList_ ) _printMessage( message ); }

//_________________________________________________
void MainWindow::_printMessage( ChatMessage message )
{

    // do not print empty messages
    if( message.text().isEmpty() )
    {
        message.setBlockRange( 0, 0 );
        ChatMessage::Set::iterator iter( messageSet_.find( message ) );
        messageSet_.erase( iter );
        messageSet_.insert( message );
        return;
    }

    // print date
    if( !(lastMessage_.isValid() && lastMessage_.receptionTime().isSameDay( message.receptionTime() ) ) )
    {
        messageBoardWidget_->appendDate( message.receptionTime() );
        lastMessage_ = ChatMessage();
    }

    // find block at which message will be appended
    const int firstBlock( messageBoardWidget_->document()->blockCount() );

    // format text
    QString text;
    if( lastMessage_.isValid() && lastMessage_.id() == message.id() )
    {
        text = QString( "[%1] ... %2" )
            .arg( message.receptionTime().toString( TimeStamp::TimeLong ) )
            .arg( message.text() );

    } else {

        text = QString( "[%1] %2: %3" )
            .arg( message.receptionTime().toString( TimeStamp::TimeLong ) )
            .arg( message.id().userString() )
            .arg( message.text() );

    }

    // append to message board
    messageBoardWidget_->appendMessage( text );

    // find last block corresponding to message
    const int lastBlock( messageBoardWidget_->document()->blockCount()-1 );

    // assign block range to message and re-insert in set.
    // this is expensive. Should try to see if there is a better way (by keeping track of the iterator)
    message.setBlockRange( firstBlock, lastBlock );
    ChatMessage::Set::iterator iter( messageSet_.find( message ) );
    messageSet_.erase( iter );
    messageSet_.insert( message );

    // update last message
    lastMessage_ = message;

}

//______________________________________________________________________
int MainWindow::_connectedClientsCount( void ) const
{

    int out( 0 );
    foreach( const auto& id, connectedClientsModel_.get() )
    { if( id.userState() != ConnectionId::Offline ) ++out; }

    return out;
}
