#ifndef MainWindow_h
#define MainWindow_h

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "BaseMainWindow.h"
#include "ChatMessage.h"
#include "Counter.h"
#include "ConnectionId.h"
#include "ConnectedClientsModel.h"
#include "ClientInterface.h"
#include "Emoticon.h"
#include "MessageBoardWidget.h"
#include "SingleTimeDialog.h"

#include <QActionGroup>
#include <QBasicTimer>
#include <QToolButton>

class ConnectedClientsList;
class HistoryMenu;
class IdleTime;
class MessageEditorWidget;
class ProgressStatusBar;

class MainWindow: public BaseMainWindow, public Counter
{

    //* Qt meta object declaration
    Q_OBJECT

    public:

    //* constructor
    MainWindow( QWidget* = nullptr );

    //* destructor
    virtual ~MainWindow( void );

    //*@name accesors
    //@{

    //* status bar
    ProgressStatusBar& statusBar( void )
    { return *statusBar_; }

    //*@name actions
    //@{

    //* connection
    QAction& toggleConnectionAction( void ) const
    { return *toggleConnectionAction_; }

    //* fetch connected clients
    QAction& fetchConnectedClientsAction( void ) const
    { return *fetchConnectedClientsAction_; }

    //* connection
    QAction& pingServerAction( void ) const
    { return *pingServerAction_; }

    //* connection
    QAction& pingClientsAction( void ) const
    { return *pingClientsAction_; }

    //* reload action
    QAction& reloadAction( void ) const
    { return *reloadAction_; }

    //* history menu
    HistoryMenu& historyMenu( void ) const
    { return *historyMenu_; }

    //* user state action group
    QActionGroup& userStateActions( void ) const
    { return *userStateActions_; }

    //* idle state
    IdleTime& idleTime( void ) const
    { return *idleTime_; }

    //@}

    //* event filter
    virtual bool eventFilter( QObject* ,QEvent* );

    Q_SIGNALS:

    //* request connection
    void requestConnection( void );

    //* request disconnection
    void requestDisconnection( void );

    //* send a new message
    void newMessage( ChatMessage );

    //* send user typing
    void userTyping( void );

    //* new notification
    void newNotification( QString );

    //* user state has changed
    void userStateChanged( ConnectionId::UserState );

    //* emitted when there are pending messages
    void hasPendingMessages( bool );

    public Q_SLOTS:

    //* id
    void setId( ConnectionId );

    //* client connection state changed
    void updateConnectionState( ConnectionId, ClientInterface::State );

    //* update idle state
    void updateIdleState( bool );

    //* update user state
    void updateUserState( ConnectionId::UserState );

    //* update user state
    void updateUserState( ConnectionId );

    //* update busy state
    void updateBusyState( bool, int );

    //* new error recieved
    void printError( QString, bool );

    //* new message recieved
    void printMessage( ChatMessage );

    //* clear list of connected clients
    void clearConnectedClients( void );

    //* clear local message history
    void clearMessageList( void );

    //* notify that a user is typing
    void notifyUserTyping( ConnectionId );

    protected:

    //* change event
    virtual void changeEvent( QEvent* );

    //* close event
    virtual void closeEvent( QCloseEvent* );

    //* timer event
    virtual void timerEvent( QTimerEvent* );

    protected Q_SLOTS:

    //* toggle connection
    void _toggleConnection( void );

    //* send message from editor
    void _sendMessage( void );

    //* user state action triggered
    void _userStateActionTriggered( QAction* );

    //* insert emoticon
    void _insertEmoticon( Emoticon );

    //* editor modified
    void _editorModified( void );

    //* edit context menu message
    void _editContextMenuMessage( void );

    //* remove context menu message
    void _removeContextMenuMessage( void );

    //* remove replace message
    void _removeReplaceMessage( void );

    //* cancel message edition
    void _cancelMessageEdition( void );

    //* view link
    void _openLink( void );

    //* view link
    void _openLink( QString );

    private Q_SLOTS:

    //* configuration
    void _updateConfiguration( void );

    private:

    //* install actions
    void _installActions( void );

    //* add message to internal storing
    /** returns true if message list order was changed */
    bool _appendMessage( const ChatMessage& );

    //* clear replace iterator
    void _clearReplaceIterator( bool clearMessage = true );

    //* update status bar
    void _updateStatusBar( void ) const;

    //* update user state button
    void _updateUserStateButton( void ) const;

    //* update replace iterator so that it matches message
    void _updateReplaceIterator( void );

    //* print entire message list
    void _printMessageList( void );

    //* print message
    void _printMessage( ChatMessage );

    //* number of connected (i.e. non Offline) clients
    int _connectedClientsCount( void ) const;

    //* id
    ConnectionId id_;

    //* idle time
    IdleTime* idleTime_ = nullptr;

    //* message board
    MessageBoardWidget* messageBoardWidget_ = nullptr;

    //* line editor
    MessageEditorWidget* messageEditorWidget_ = nullptr;

    //* status bar
    ProgressStatusBar* statusBar_ = nullptr;

    //* connected state
    ClientInterface::State state_ = ClientInterface::Disconnected;

    //* connected clients list
    ConnectedClientsList* connectedClientsList_ = nullptr;

    //* connected clients model
    ConnectedClientsModel connectedClientsModel_;

    //* chat messages list
    ChatMessage::Set messageSet_;
    ChatMessage::List messageList_;

    //* message replacement
    ChatMessage::List::iterator replaceIterator_;
    ChatMessage replaceMessage_;

    //* last chat message
    ChatMessage lastMessage_;

    //* message that matches context menu event
    ChatMessage contextMenuMessage_;

    //*@name actions
    //@{

    //* toggle connection
    QAction* toggleConnectionAction_ = nullptr;

    //* fetch clients
    QAction* fetchConnectedClientsAction_ = nullptr;

    //* ping server
    QAction* pingServerAction_ = nullptr;

    //* ping clients
    QAction* pingClientsAction_ = nullptr;

    //* history action
    QToolButton* historyButton_ = nullptr;

    //* user state button
    QToolButton* userStateButton_ = nullptr;

    //* history menu
    HistoryMenu* historyMenu_ = nullptr;

    //* user state actions
    QActionGroup* userStateActions_ = nullptr;

    //* reload action
    QAction* reloadAction_ = nullptr;

    //* map actions to delay
    using ActionMap = QMap<QAction*, ConnectionId::UserState>;
    ActionMap userStateActionMap_;

    //@}

    //* true if progress bar must be shown when fetching logs
    bool showFetchProgress_ = false;

    //* true if mainwindow is busy (recieving logs)
    /** no notifications will be sent */
    bool busy_ = false;

    //*@name user typing
    //@{

    //* user typing timeout
    int userTypingTimeout_ = 500;

    //* this user typing
    QBasicTimer thisUserTypingTimer_;

    //* user typing timer
    QBasicTimer userTypingTimer_;

    //@}

};

#endif
