/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "MessageBoardWidget.h"

#include "BaseContextMenu.h"
#include "BaseIconNames.h"
#include "EmoticonProvider.h"
#include "File.h"
#include "HtmlHighlighter.h"
#include "IconEngine.h"
#include "QtUtil.h"
#include "SyntaxHighlighter.h"
#include "XmlOptions.h"

#include <QMimeData>

//______________________________________________________
MessageBoardWidget::MessageBoardWidget( QWidget* parent ):
    TextEditor( parent )
{
    Debug::Throw( "MessageBoardWidget::MessageBoardWidget.\n" );

    setReadOnly( true );
    setTrackAnchors( true );

    // disable wrapping option
    setWrapFromOptions( false );
    wrapModeAction().setChecked( true );
    wrapModeAction().setEnabled( false );

    connect( verticalScrollBar(), SIGNAL(sliderMoved(int)), SLOT(_verticalScrollBarMoved(int)) );

    // syntax highlighting
    syntaxHighlighter_ = new SyntaxHighlighter( document() );
    htmlHighlighter_ = new HtmlHighlighter( this );
    emoticonProvider_ = new EmoticonProvider( this );
    emoticonProvider_->installResources( document() );

    // default formats
    const QTextCursor cursor( textCursor() );
    defaultBlockFormat_ = QTextBlockFormat( cursor.blockFormat() );
    defaultCharFormat_ = QTextCharFormat( cursor.charFormat() );

    // actions
    _installActions();

}

//_____________________________________________________________
void MessageBoardWidget::appendDate( const TimeStamp& time )
{

    // margin
    const int margin( fontMetrics().height()/2 );

    // move cursor to end of document
    QTextCursor cursor( textCursor() );
    cursor.movePosition( QTextCursor::End );

    // restore default and insert new block
    if( !document()->isEmpty() )
    {
        cursor.setCharFormat( defaultCharFormat_ );
        cursor.insertBlock();
    }

    // block format
    QTextBlockFormat blockFormat;
    blockFormat.setAlignment( Qt::AlignHCenter );
    blockFormat.setTopMargin( margin );
    blockFormat.setBottomMargin( margin );
    cursor.setBlockFormat( blockFormat );

    // char format
    QTextCharFormat charFormat;
    charFormat.setFont( QtUtil::titleFont( charFormat.font() ) );
    charFormat.setFontUnderline( true );

    // disable updates if locked
    if( locked_ ) setUpdatesEnabled( false );

    // convert timestamp to QDate, then text
    const QDate date( QDateTime::fromTime_t( time ).date() );
    const QString text( date.toString( Qt::SystemLocaleLongDate ) );
    cursor.insertText( text, charFormat );

    // update
    if( locked_ ) setUpdatesEnabled( true );
    else {

        setTextCursor( cursor );
        ensureCursorVisible();

    }

    return;

}

//______________________________________________________
void MessageBoardWidget::appendMessage( const QString& text )
{

    // move cursor to end of document
    QTextCursor cursor( document()->lastBlock() );
    cursor.movePosition( QTextCursor::End );

    // restore default and insert new block
    if( !document()->isEmpty() )
    {
        cursor.setCharFormat( defaultCharFormat_ );
        cursor.insertBlock();
        cursor.setBlockFormat( defaultBlockFormat_ );
    }

    // disable updates if locked
    if( locked_ ) setUpdatesEnabled( false );

    cursor.insertText( text );

    // html links
    htmlHighlighter_->process( QTextCursor( document()->lastBlock() ) );

    // emoticons
    emoticonProvider_->process( QTextCursor( document()->lastBlock() ) );

    // update
    if( locked_ ) setUpdatesEnabled( true );
    else {

        setTextCursor( cursor );
        ensureCursorVisible();

    }

    return;

}

//______________________________________________________________________________
void MessageBoardWidget::installContextMenuActions( BaseContextMenu* menu, bool allActions )
{

    Debug::Throw( "MessageBoardWidget::installContextMenuActions.\n" );

    // retrieve default context menu
    // second argument is to remove un-necessary actions
    TextEditor::installContextMenuActions( menu, allActions );

    // open link
    if( !anchor().isEmpty() )
    { menu->insertAction( &copyLinkAction(), openLinkAction_ ); }

    // add specific actions
    menu->addSeparator();
    menu->addAction( editMessageAction_ );
    menu->addAction( removeMessageAction_ );

    return;
}

//______________________________________________________
void MessageBoardWidget::clear( void )
{
    BaseEditor::clear();
    emoticonProvider_->installResources( document() );
}

//______________________________________________________
void MessageBoardWidget::_verticalScrollBarMoved( int value )
{ locked_ = value != verticalScrollBar()->maximum(); }

//______________________________________________________
void MessageBoardWidget::keyPressEvent( QKeyEvent* event )
{

    switch( event->key() )
    {
        case Qt::Key_PageUp:
        case Qt::Key_PageDown:
        case Qt::Key_Up:
        case Qt::Key_Down:
        case Qt::Key_Home:
        case Qt::Key_End:
        locked_ = verticalScrollBar()->value() != verticalScrollBar()->maximum();
        break;

        default: break;
    }

    return TextEditor::keyPressEvent( event );

}

//______________________________________________________
void MessageBoardWidget::wheelEvent( QWheelEvent* event )
{
    locked_ = verticalScrollBar()->value() != verticalScrollBar()->maximum();
    return TextEditor::wheelEvent( event );
}

//______________________________________________________
void MessageBoardWidget::changeEvent( QEvent* event )
{

    switch( event->type() )
    {
        default: break;

        case QEvent::EnabledChange:
        syntaxHighlighter_->setEnabled( isEnabled() );
        syntaxHighlighter_->setDocument( document() );
        break;
    }

    // parent class
    return TextEditor::changeEvent( event );

}

//______________________________________________________
QMimeData* MessageBoardWidget::createMimeDataFromSelection( void ) const
{

    // get mime data from parent method
    auto mimeData( TextEditor::createMimeDataFromSelection() );

    // do nothing if mimeData is invalid or emoticon provider is disabled
    if( !mimeData ) return mimeData;
    if( !emoticonProvider_->isEnabled() ) return mimeData;

    // get selection from cursor
    QTextCursor cursor( textCursor() );
    if( !cursor.hasSelection() ) return mimeData;

    // 'unparse' icons into emoticon string
    bool changed( false );
    QString text( emoticonProvider_->revert( cursor, changed ) );
    if( !changed ) return mimeData;

    // copy mime type and data
    QMimeData* copy( new QMimeData() );
    foreach( const auto& format, mimeData->formats() )
    { copy->setData( format, mimeData->data( format ) ); }

    copy->setText( text );

    // delete old, return new
    delete mimeData;
    return copy;

}


//_________________________________________________
void MessageBoardWidget::_installActions( void )
{

    Debug::Throw( "MessageBoardWidget::_installActions.\n" );

    // open link
    addAction( openLinkAction_ = new QAction( IconEngine::get( IconNames::Find ), tr( "Open Link..." ), this ) );

    // edit messsage actions
    addAction( editMessageAction_ = new QAction( IconEngine::get( IconNames::Edit ), tr( "Edit Message" ), this ) );

    // remove message actions
    addAction( removeMessageAction_ = new QAction( IconEngine::get( IconNames::Cut ), tr( "Remove Message" ), this ) );
    removeMessageAction_->setShortcut( QKeySequence::Cut );

}
