/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ServerInterface.h"

#include "Singleton.h"
#include "XmlDocument.h"
#include "XmlDef.h"
#include "XmlOptions.h"

#include <QCoreApplication>
#include <QDateTime>
#include <QHostInfo>

//____________________________________________
ServerInterface::ServerInterface( QObject* parent ):
    QObject( parent ),
    Counter( "ServerInterface" )
{

    // configuration
    connect( Singleton::get().application(), SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    connect( qApp, SIGNAL(aboutToQuit()), SLOT(_saveConfiguration()) );
    _updateConfiguration();
}

//____________________________________________
ServerInterface::~ServerInterface( void ) = default;

//____________________________________________
void ServerInterface::initialize( const CommandLineParser& parser )
{
    Debug::Throw( "ServerInterface::initialize.\n" );

    // store parser
    parser_ = parser;

    if( !server_ )
    {
        // create server
        server_ = new QTcpServer( this );
        connect( server_, SIGNAL(newConnection()), SLOT(_newConnection()) );
    }

    // host and port from arguments
    if( parser.hasOption( "--chat-host" ) ) host_ = QHostAddress( parser.option( "--chat-host" ) );
    if( parser.hasOption( "--chat-port" ) ) port_ = parser.option( "--chat-port" ).toUInt();

    // initialize
    Debug::Throw() << "ServerInterface::initialize - Listening to " << host_.toString() << ":" << port_ << endl;

    if( !server_->listen( host_, port_ ) )
    {  Debug::Throw(0) << "ServerInterface::initialize - Failed to listen to " << host_.toString() << ":" << port_ << endl; }

}

//____________________________________________
bool ServerInterface::readMessages()
{
    Debug::Throw(0, "ServerInterface::readMessages.\n" );
    if( dbFile_.isEmpty() )
    {
        Debug::Throw(0) << "ServerInterface::readMessages - invalid message list." << endl;
        return false;
    }

    Debug::Throw(0) << "ServerInterface::readMessages - db file: " << dbFile_ << endl;

    // parse the file
    XmlDocument document;
    if( !_readXmlDocument( document ) )
    { return false; }

    // look for relevant element
    QDomNodeList topNodes = document.elementsByTagName( Xml::MessageList );
    if( topNodes.isEmpty() )
    {
        Debug::Throw() <<  "ServerInterface::readMessages - top node not found" << endl;
        return false;

    }

    TimeStamp timeStamp;
    switch( timeSpan_ )
    {
        case OneWeek: timeStamp = QDateTime(  QDate::currentDate() ).addDays(-7).toTime_t(); break;
        case TwoWeeks: timeStamp = QDateTime(  QDate::currentDate() ).addDays(-14).toTime_t(); break;
        case OneMonth: timeStamp = QDateTime( QDate::currentDate() ).addMonths(-1).toTime_t(); break;
        case TwoMonths: timeStamp = QDateTime( QDate::currentDate() ).addMonths(-2).toTime_t(); break;
        case SixMonths: timeStamp = QDateTime( QDate::currentDate() ).addMonths(-6).toTime_t(); break;
        case OneYear: timeStamp = QDateTime( QDate::currentDate() ).addYears(-1).toTime_t(); break;
        default: break;
    }

    if( !timeStamp.isValid() ) messageSet_ = ChatMessage::Set( topNodes.at(0).toElement() );
    else {

        // read messages from log and truncate
        messageSet_.clear();
        const ChatMessage::Set messageSet( ChatMessage::Set( topNodes.at(0).toElement() ) );
        foreach( const auto& message, messageSet )
        {
            if( !message.creationTime().isValid() || message.creationTime() > timeStamp )
            { messageSet_.insert( message ); }
        }

    }

    Debug::Throw(0) << "ServerInterface::readMessages - size: " << messageSet_.size() << endl;
    return true;

}

//____________________________________________
bool ServerInterface::writeMessages()
{
    Debug::Throw(0, "ServerInterface::writeMessages.\n" );
    if( dbFile_.isEmpty() )
    {
        Debug::Throw(0) << "ServerInterface::writeMessages - invalid message list." << endl;
        return false;
    }

    // create document and read
    XmlDocument document;
    _readXmlDocument( document );

    // append truncated list
    QDomElement top( messageSet_.domElement( document ) );
    document.replaceChild( top );
    {
        QFile qfile( dbFile_ );
        if( !qfile.open( QIODevice::WriteOnly ) )
        {

            // write log
            Debug::Throw(0) << "ServerInterface::writeMessages - cannot open file: " << dbFile_ << " for writting." << endl;

            // send a message to all clients
            ChatCommand command( ConnectionId( "server" ), ChatCommand::Message );
            command.setCreationTime( TimeStamp::now() );
            command.setReceptionTime( TimeStamp::now() );
            command.setText( tr( "Unable to write messages to disc." ) );
            _broadcast( command );

            command.setCreationTime( TimeStamp::now()+1 );
            command.setReceptionTime( TimeStamp::now()+1 );
            command.setText( tr( "Please check write permissions on server host." ) );
            _broadcast( command );

            return false;
        }

        if( useCompression_ ) qfile.write( qCompress( document.toByteArray() ) );
        else qfile.write( document.toByteArray() );
    }

    return true;

}

//_____________________________________________________
void ServerInterface::timerEvent( QTimerEvent* event )
{
    if( event->timerId() == autoSaveTimer_.timerId() )
    {
        autoSaveTimer_.stop();
        writeMessages();
    } else return QObject::timerEvent( event );
}

//_____________________________________________________
void ServerInterface::_newConnection()
{
    Debug::Throw( 0, "ServerInterface::_newConnection.\n" );

    // check pending connection
    if( !server_->hasPendingConnections() ) return;

    // create client from pending connection
    QTcpSocket* socket( server_->nextPendingConnection() );
    ChatClient *client( new ChatClient( this, socket ) );
    connect( &client->socket(), SIGNAL(disconnected()), SLOT(_clientDisconnected()) );
    connect( client, SIGNAL(commandAvailable(ChatCommand)), SLOT(_processCommand(ChatCommand)) );
    connectedClients_.append( client );

}

//_____________________________________________________
void ServerInterface::_clientDisconnected( void )
{

    // remove from accepted client map
    {
        ClientMap::iterator iter;
        while( ( iter = std::find_if(  acceptedClients_.begin(), acceptedClients_.end(), SameStateFTor( QAbstractSocket::UnconnectedState ) )  ) != acceptedClients_.end() )
        {
            Debug::Throw(0) << "ServerInterface::_clientDisconnected - user:" << iter.key().user() << endl;

            // update id status
            ConnectionId id( iter.key() );
            id.setUserState( ConnectionId::Offline );
            id.setDisconnectionTime( TimeStamp::now() );

            // broadcast client has disconnected and status changed
            _broadcast( ChatCommand( id, ChatCommand::UserStateChanged ), iter.value() );
            _broadcast( ChatCommand( id, ChatCommand::ClientDisconnected ), iter.value() );

            // erase from map of accepted clients
            acceptedClients_.erase( iter );

        }
    }

    // remove from list of connected clients
    {
        ClientList::iterator iter;
        while( ( iter = std::find_if( connectedClients_.begin(), connectedClients_.end(), ChatClient::SameStateFTor( QAbstractSocket::UnconnectedState ) ) ) != connectedClients_.end() )
        {
            (*iter)->deleteLater();
            connectedClients_.erase( iter );
        }
    }

}

//_____________________________________________________
void ServerInterface::_processCommand( ChatCommand command )
{

    Debug::Throw(0)
        << "ServerInterface::_processCommand -"
        << " time: " << command.creationTime().toString()
        << " user: " << command.id().user()
        << " command: " << command.commandName()
        << endl;

    ClientList::iterator iter( std::find_if(  connectedClients_.begin(), connectedClients_.end(), ChatClient::SameIdFTor( command.clientId() ) ) );
    Q_ASSERT( iter != connectedClients_.end() );
    _processCommand( command, *iter );

    return;
}

//_____________________________________________________
void ServerInterface::_processCommand( ChatCommand command, ChatClient* sender )
{

    switch( command.command() )
    {

        // connection requested by client
        case ChatCommand::ConnectionRequest:
        {
            // register to accepted clients
            ConnectionId id( command.id() );
            id.setConnectionTime( TimeStamp::now() );
            ClientMap::iterator clientIterator( _register( id, sender ) );

            // return accept to calling client
            sender->sendCommand( ChatCommand( id, ChatCommand::ConnectionAccepted ) );

            // broadcast client connected and UserStateChanged
            _broadcast( ChatCommand( id, ChatCommand::ClientConnected ), sender );
            _broadcast( ChatCommand( id, ChatCommand::UserStateChanged ), sender );
            break;
        }

        // ping request from client
        case ChatCommand::PingServer:
        {
            sender->sendCommand( ChatCommand( command.id(), ChatCommand::PongServer ) );
            break;
        }

        // message sent
        case ChatCommand::Message:
        {

            // append to local message list
            ChatMessage message( command );

            // set recepion time stamp if not already done
            if( !message.receptionTime().isValid() )
            { message.setReceptionTime( TimeStamp::now() ); }

            _appendMessage( message );

            // trigger autosave
            if( autoSaveTimeout_ > 0 )
            {

                if( !autoSaveTimer_.isActive() ) autoSaveTimer_.start( autoSaveTimeout_, this );

            } else writeMessages();

            // broadcast to all connected clients but sender
            _broadcast( command, sender );
            break;

        }

        // message list request
        case ChatCommand::FetchMessages:
        {

            // find start time
            TimeStamp timeStamp;
            if( chatCommandLineParser_.parse( command.arguments() ).hasOption( "-t" ) )
            {
                timeStamp.setTime( chatCommandLineParser_.option( "-t" ).toInt() );
                Debug::Throw(0) << "ServerInterface::_processCommand - timeStamp: " << timeStamp.toString() << endl;
            }

            // find all matching messages
            ChatMessage::List messages;
            ChatMessage last;
            foreach( const auto& message, messageSet_ )
            {
                if( !last.isValid() || message.receptionTime() > last.receptionTime() )
                { last = message; }

                // skip message if too old
                if( timeStamp.isValid() && message.receptionTime() < timeStamp ) continue;

                // store in list
                messages.append( message );

            }

            if( messages.isEmpty() && last.isValid() ) messages.append( last );
            if( messages.isEmpty() ) break;

            // sort messages
            qSort( messages.begin(), messages.end(), ChatMessage::SortReceptionTimeFTor() );

            // notify start message log
            {
                ChatCommand copy( command.id(), ChatCommand::StartMessageLog );
                copy.setArguments( QStringList() << copy.commandName() << "-s" << QString::number( messages.size() ) );
                sender->sendCommand( copy );
            }

            // clear message list
            sender->sendCommand( ChatCommand( command.id(), ChatCommand::ClearMessageList ) );

            Debug::Throw(0) << "ServerInterface::_processCommand - found " << messages.size() << " messages" << endl;

            // send messages
            foreach( const auto& message, messages )
            { sender->sendCommand( message.toChatCommand() ); }

            // notify end message log
            sender->sendCommand( ChatCommand( command.id(), ChatCommand::StopMessageLog ) );

            break;
        }

        // connected client list request
        case ChatCommand::FetchConnectedClients:
        {

            // clear list of connected clients
            sender->sendCommand( ChatCommand( command.id(), ChatCommand::ClearConnectedClients ) );

            // loop over connected clients
            for( auto iter = acceptedClients_.begin(); iter != acceptedClients_.end(); ++iter )
            { sender->sendCommand( ChatCommand( iter.key(), ChatCommand::ClientConnected ) ); }

            break;

        }

        // user state changed
        case ChatCommand::UserStateChanged:
        {
            // change status stored in accepted clients
            ClientMap::iterator iter( acceptedClients_.find( command.id() ) );
            if( iter != acceptedClients_.end() )
            {
                ConnectionId& id( const_cast<ConnectionId&>( iter.key() ) );
                id.setUserState( command.id().userState() );
                command.setId( id );
            }

            // broadcast
            _broadcast( command, sender );
            break;

        }

        // user state changed
        case ChatCommand::UsernameChanged:
        {

            // find matching iterator in accepted clients list
            auto iter = std::find_if( acceptedClients_.begin(), acceptedClients_.end(), SameClientFTor(sender) );
            if( iter == acceptedClients_.end() ) break;

            // check if user name indeed changed
            ConnectionId id( iter.key() );
            if( id.user() == command.id().user() ) break;

            // emit user disconnected
            {
                ConnectionId local( id );
                local.setUserState( ConnectionId::Renamed );
                _broadcast( ChatCommand( local, ChatCommand::UserStateChanged ), sender );
                _broadcast( ChatCommand( local, ChatCommand::ClientDisconnected ), sender );
            }

            // remove from list
            acceptedClients_.erase( iter );

            // reinsert with new user
            id.setUser( command.id().user() );
            acceptedClients_.insert( id, sender );

            // broadcast client connected and UserStateChanged
            _broadcast( ChatCommand( id, ChatCommand::ClientConnected ), sender );
            _broadcast( ChatCommand( id, ChatCommand::UserStateChanged ), sender );

            break;
        }

        // all other commands, just broadcast
        default:
        case ChatCommand::PingClient:
        case ChatCommand::PongClient:
        {
            _broadcast( command, sender );
            break;
        }
    }

    return;

}

//_____________________________________________________
void ServerInterface::_broadcast( ChatCommand command, ChatClient* sender )
{

    Debug::Throw() << "ServerInterface::_Broadcast - id: " << command.id().user() << " command: " << command.commandName() << endl;
    for( ClientList::iterator iter = connectedClients_.begin(); iter != connectedClients_.end(); ++iter )
    { if( (*iter) != sender ) (*iter)->sendCommand( command ); }

}

//_____________________________________________________
ServerInterface::ClientMap::iterator ServerInterface::_register( const ConnectionId& id, ChatClient* client )
{
    Debug::Throw( "ServerInterface::_register.\n" );

    ClientMap::iterator iter( acceptedClients_.find( id ) );
    if( iter != acceptedClients_.end() )
    {

        // check if client is the same
        if( iter.value() == client ) return iter;
        else {

            iter.value()->socket().disconnectFromHost();
            acceptedClients_.erase( iter );
            return acceptedClients_.insert( id, client );

        }


    } else {

        // see if there is already a client in the list
        if( (iter = std::find_if( acceptedClients_.begin(), acceptedClients_.end(), SameClientFTor(client) ) ) != acceptedClients_.end() )
        { acceptedClients_.erase( iter ); }

        return acceptedClients_.insert( id, client );

    }

}

//______________________________________
void ServerInterface::_updateConfiguration( void )
{
    Debug::Throw( "ServerInterface::_updateConfiguration.\n" );

    // hostName
    if( !parser_.hasOption( "--chat-host" ) )
    {
        QHostInfo hostInfo( QHostInfo::fromName( XmlOptions::get().raw( "CHAT_HOST" ) ) );
        if( !hostInfo.addresses().isEmpty() ) host_ = hostInfo.addresses().front();
    }

    // port
    if( !parser_.hasOption( "--chat-port" ) )
    { port_ = XmlOptions::get().get<unsigned int>( "CHAT_PORT" ); }

    // compression
    useCompression_ = XmlOptions::get().get<bool>( "COMPRESS_DB_FILE" );

    // log timespan
    timeSpan_ = (TimeSpan) XmlOptions::get().get<int>( "LOG_SPAN" );

    // DB file
    File dbFile( XmlOptions::get().raw("DB_FILE") );
    if( dbFile != dbFile_ )
    {
        dbFile_ = dbFile;
        readMessages();
    }

    // auto save timeout
    autoSaveTimeout_ = 1000*XmlOptions::get().get<int>( "AUTOSAVE_TIMEOUT" );
    Debug::Throw() << "ServerInterface::_updateConfiguration - autosave: " << autoSaveTimeout_ << endl;

    return;

}

//______________________________________
void ServerInterface::_saveConfiguration( void )
{
    Debug::Throw( "ServerInterface::_saveConfiguration.\n" );
    writeMessages();
}

//_________________________________________________
void ServerInterface::_appendMessage( ChatMessage message )
{

    // first check if message is in set
    ChatMessage::Set::iterator iter( messageSet_.find( message ) );
    if( iter != messageSet_.end() )
    {
        message.setReceptionTime( iter->receptionTime() );
        messageSet_.erase( iter );
    }
    messageSet_.insert( message );
    return;

}

//_________________________________________________
bool ServerInterface::_readXmlDocument( XmlDocument& document )
{

    Debug::Throw( "ServerInterface::_readXmlDocument.\n" );
    QFile qfile( dbFile_ );
    if( !qfile.open( QIODevice::ReadOnly ) )
    {
        Debug::Throw(0) << "ServerInterface::_readXmlDocument - cannot open" << endl;
        return false;
    }

    if( useCompression_ )
    {

        Debug::Throw( "ServerInterface::_readXmlDocument - trying compressed.\n" );
        // try read compressed and try uncompress
        QByteArray content( qfile.readAll() );
        QByteArray uncompressed = qUncompress( content );

        // try read raw if failed
        if( !uncompressed.size() ) uncompressed = content;

        // pass contents to XmlDocument, after trying to uncompress
        if( !document.setContent( uncompressed ) )
        {
            Debug::Throw(0) << document.error() << endl;
            return false;
        }

    } else {

        // read raw
        Debug::Throw( "ServerInterface::_readXmlDocument - trying uncompressed.\n" );
        QByteArray uncompressed( qfile.readAll() );
        bool success( false );
        if( !( success = document.setContent( uncompressed ) ) )
        {
            // try read compressed and uncompress
            QByteArray content( qfile.readAll() );
            QByteArray uncompressed = qUncompress( content );
            success = document.setContent( uncompressed );
        }

        // pass contents to XmlDocument, after trying to uncompress
        if( !success )
        {
            Debug::Throw(0) << document.error() << endl;
            return false;
        }

    }

    return true;

}

//____________________________________________________
QTextStream& operator << (QTextStream& out, const ServerInterface::ClientList& clientList )
{
    foreach( const auto& client, clientList )
    { out << "Client: " << client <<  endl; }
    return out;
}

//____________________________________________________
QTextStream& operator << (QTextStream& out, const ServerInterface::ClientMap& clientMap )
{
    for( auto iter = clientMap.constBegin(); iter != clientMap.constEnd(); ++iter )
    {  out << "Id: " << iter.key() << " " << iter.value() << endl; }
    return out;
}
