#ifndef ServerInterface_h
#define ServerInterface_h

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ChatClient.h"
#include "ChatCommandLineParser.h"
#include "ChatMessage.h"
#include "File.h"
#include "XmlDocument.h"

#include <QBasicTimer>
#include <QHostAddress>
#include <QTcpServer>
#include <QTimerEvent>

class ServerInterface: public QObject, public Counter
{

    //* Qt meta object declaration
    Q_OBJECT

    public:

    //* constructor
    ServerInterface( QObject* );

    //* destructor
    virtual ~ServerInterface( void );

    //*@name accessors
    //@{

    const ChatMessage::Set& messages( void ) const
    { return messageSet_; }

    //@}

    //*@name modifiers
    //@{

    //* initialize
    void initialize( const CommandLineParser& );

    //* host
    void setHost( QHostAddress value )
    { host_ = value; }

    //* port
    void setPort( int value )
    { port_ = value; }

    //* set messages
    void setMessages( const ChatMessage::Set& messages )
    { messageSet_ = messages; }

    //* read messages
    bool readMessages( void );

    //* save messages
    bool writeMessages( void );

    //@}

    //* history span
    enum TimeSpan
    {
        OneWeek,
        TwoWeeks,
        OneMonth,
        TwoMonths,
        SixMonths,
        OneYear,
        Forever
    };

    protected Q_SLOTS:

    //* a new connection is granted
    virtual void _newConnection( void );

    //* client disconnected
    virtual void _clientDisconnected( void );

    //* command processing
    virtual void _processCommand( ChatCommand );

    protected:

    //* timer event
    void timerEvent( QTimerEvent* );

    //* pair of application id and client
    using ClientPair = QPair< ConnectionId, ChatClient* >;

    //* list of clients
    using ClientList = QList< ChatClient* >;

    //* map of clients
    using ClientMap = QMap< ConnectionId, ChatClient* >;

    //* used to retrieve clients for a given state
    class SameStateFTor: public ChatClient::SameStateFTor
    {
        public:

        //* constructor
        SameStateFTor( QAbstractSocket::SocketState state ):
            ChatClient::SameStateFTor( state )
        {}

        //* destructor
        virtual ~SameStateFTor( void )
        {}

        //* predicate
        bool operator() ( const ClientPair& pair ) const
        { return ChatClient::SameStateFTor::operator() (pair.second); }

        //* predicate
        bool operator() ( const ChatClient* client ) const
        { return ChatClient::SameStateFTor::operator() (client); }

    };

    //* used to retrieve pair with matching client
    class SameClientFTor
    {
        public:

        //* constructor
        SameClientFTor( ChatClient* client ):
            client_( client )
        {}

        //* destructor
        virtual ~SameClientFTor( void )
        {}

        //* predicate
        bool operator() ( const ClientPair& pair ) const
        { return pair.second == client_; }

        //* predicate
        bool operator() ( ChatClient* client ) const
        { return client == client_; }

        private:

        //* prediction
        ChatClient* client_;
    };

    //* process command
    void _processCommand( ChatCommand, ChatClient* );

    //* broadcast a message to all registered clients but the sender (if valid)
    virtual void _broadcast( ChatCommand, ChatClient* sender = 0 );

    //* register a client
    virtual ClientMap::iterator _register( const ConnectionId& id, ChatClient* client );

    private Q_SLOTS:

    //* update configuration
    void _updateConfiguration( void );

    //* save configuration
    void _saveConfiguration( void );

    private:

    //* read Xml document from QFile
    bool _readXmlDocument( XmlDocument& );

    //* add message to internal storing
    void _appendMessage( ChatMessage );

    // command line parser (passed at initialization)
    CommandLineParser parser_;

    // host
    QHostAddress host_ = QHostAddress::Null;

    // port
    unsigned int port_ = 0;

    //* Server
    QTcpServer* server_ = nullptr;

    //* chat command line parser
    ChatCommandLineParser chatCommandLineParser_;

    //* chat message list
    ChatMessage::Set messageSet_;

    //* list of connected clients
    ClientList connectedClients_;

    //* maps accepted clients and amount of request
    ClientMap acceptedClients_;

    //* file from/to wich the files are saved
    File dbFile_;

    //* compression
    bool useCompression_ = false;

    //* time span
    TimeSpan timeSpan_ = Forever;

    //*@name autosave
    //@{

    QBasicTimer autoSaveTimer_;
    int autoSaveTimeout_ = 0;

    //@}

    //* streamers
    friend QTextStream& operator << (QTextStream&, const ClientList& );
    friend QTextStream& operator << (QTextStream&, const ClientMap& );


};

#endif
