/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "SystemTrayIcon.h"

#include "ClientApplication.h"
#include "IconEngine.h"
#include "Singleton.h"
#include "SystemTrayMenu.h"

#include <QApplication>
#include <QDesktopWidget>

//______________________________________________________
SystemTrayIcon::SystemTrayIcon( QObject* parent ):
    QSystemTrayIcon( parent ),
    Counter( "SystemTrayIcon" )
{

    Debug::Throw( "SystemTrayIcon::SystemTrayIcon.\n" );

    // store icons
    connectedIcon_ = static_cast<const ClientApplication*>( Singleton::get().application() )->applicationIcon();
    foreach( const auto& size, connectedIcon_.availableSizes() )
    { disconnectedIcon_.addPixmap( connectedIcon_.pixmap( size, QIcon::Disabled ) ); }

    pendingMessagesIcon_ = IconEngine::get( "messages-pending" );

    // use application icon
    setIcon( disconnectedIcon_ );

    connect( this, SIGNAL(activated(QSystemTrayIcon::ActivationReason)), this, SLOT(_activated(QSystemTrayIcon::ActivationReason)) );

    // context menu
    contextMenu_ = new SystemTrayMenu();

}

//______________________________________________________
SystemTrayIcon::~SystemTrayIcon( void )
{ contextMenu_->deleteLater(); }

//______________________________________________________
void SystemTrayIcon::updatePendingMessages( bool hasPendingMessages )
{ setIcon( hasPendingMessages ? pendingMessagesIcon_:connectedIcon_ ); }

//______________________________________________________
void SystemTrayIcon::updateConnectionState( ClientInterface::State state )
{
    switch( state )
    {

        default:
        case ClientInterface::Disconnected:
        case ClientInterface::AwaitingReply:
        setIcon( disconnectedIcon_ );
        break;

        case ClientInterface::Connected:
        setIcon( connectedIcon_ );
        break;

    }

}

//______________________________________________________
void SystemTrayIcon::_activated( QSystemTrayIcon::ActivationReason reason )
{
    switch( reason )
    {

        default: break;
        case Context: _showContextMenu(); break;
    }

}

//______________________________________________________
void SystemTrayIcon::_showContextMenu( void ) const
{

    Debug::Throw( "SystemTrayIcon::_showContextMenu.\n" );

    // icon geometry
    const QRect geometry( this->geometry() );

    // menu size
    contextMenu_->adjustSize();
    const QSize size( contextMenu_->size() );

    // desktop size
    QDesktopWidget* desktop( qApp->desktop() );
    QRect desktopGeometry( desktop->availableGeometry( desktop->screenNumber( contextMenu_ ) ) );

    // detect the side of the systray icon by comparing geometry and desktop geometry
    enum DesktopSide
    {
        Left,
        Top,
        Right,
        Bottom
    };

    DesktopSide desktopSide( Bottom );
    if( geometry.bottom() < desktopGeometry.top() ) desktopSide = Top;
    else if( geometry.right() < desktopGeometry.left() ) desktopSide = Left;
    else if( geometry.left() > desktopGeometry.right() ) desktopSide = Right;

    // define menu geometry
    int top(0);
    int left(0);

    switch( desktopSide )
    {
        case Left:
        {
            left = desktopGeometry.left();
            top = qMin( geometry.top(), desktopGeometry.bottom() - size.height() );
            break;
        }

        case Top:
        {
            top = desktopGeometry.top();
            left = qMin( geometry.left(), desktopGeometry.right() - size.width() );
            break;
        }

        case Right:
        {
            left = desktopGeometry.right() - size.width();
            top = qMin( geometry.top(), desktopGeometry.bottom() - size.height() );
            break;
        }

        default:
        case Bottom:
        {
            top = desktopGeometry.bottom() - size.height();
            left = qMin( geometry.left(), desktopGeometry.right() - size.width() );
            break;
        }

    }

    // move menu
    contextMenu_->exec(QPoint( left, top ));

}
